# run with : python3 -m unittest TestModels.TestModels.testGaussModel

import unittest
import os
import numpy as numpy
from astropy import units
import matplotlib.pyplot as plt
import warnings

from StdTests import stdModeltest

from BayesicFitting import *


__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestModels( unittest.TestCase ):
    """
    Test harness for Models

    Author:      Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    def testToString( self ) :
        model = PadeModel( 1, 2 )
        model += PolynomialModel( 0 )
        name = model.__str__()
        print( name )
        self.assertTrue( name[-1] == '4' )

        model = PadeModel( 1, 2 )
        model += PolynomialModel( 1 )
        name = str( model )
        print( name )
        self.assertTrue( name[-1] == 'x' )
        self.assertTrue( name[-5] == '5' )
        sn = model.shortName()
        print( sn )
        self.assertTrue( sn[-1] == 'l' )

    def testArctanModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******ARCTAN***********************" )
        m = ArctanModel( )
        p = numpy.asarray( [1.2,-0.1,30], dtype=float )

        stdModeltest( m, p, plot=self.doplot )

    def testArctanFixedModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******ARCTAN FIXED******************" )
        m = ArctanModel( fixed={1:-1.2} )
        p = numpy.asarray( [1.2,30], dtype=float )

        stdModeltest( m, p, plot=self.doplot )

    def testConstantModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******NULL***********************" )
        m = ConstantModel( )
        self.assertTrue( m.npchain == 0 )
        p = []
        stdModeltest( m, p )

        print( "******CONSTANT VALUE*******************" )
        sm = SineModel()
        m = ConstantModel( fixedModel=sm )
        m.values = [2.0, 0.0, 0.5]
        stdModeltest( m, p, silent=False )

        print( "******CONSTANT MODEL*******************" )
        tbl = numpy.linspace( 0, 1, 11 )
        m = ConstantModel( table=tbl )
        stdModeltest( m, p )

        print( "******CONSTANT TABLE*******************" )
        m = ConstantModel( values = 5.5 )
        stdModeltest( m, p )

        print( "******COMPOUND*******************" )
        m = ConstantModel( values = 1.0 )
        m.addModel( ExpModel() )
        self.assertTrue( m.npbase == 0 )
        self.assertTrue( m.npchain == 2 )
        p = [1.0, -3.0]
        stdModeltest( m, p, plot=self.doplot )

    def testEtalonModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******ETALON***********************" )
        m = EtalonModel( )
        p = numpy.asarray( [1.2, 0.6, 2.0, 0.2], dtype=float )

        stdModeltest( m, p, plot=self.doplot )

    """
    def testEtalonVarModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******ETALON VAR***********************" )
        m = EtalonVarModel( )
        p = numpy.asarray( [1.2, 0.6, 2.0, 0.2], dtype=float )

        stdModeltest( m, p, plot=self.doplot )

    def testEtalonVarModel2( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******ETALON VAR 2***********************" )
        pm = BasicSplinesModel( knots=[-1.0, 0.1, 1.0] )
        m = EtalonVarModel( fixed={1:pm} )
        p = numpy.asarray( [1.0, 5.5, 0.2, 0.03, 0.04, 0.05, 0.02, 0.01], dtype=float )

        stdModeltest( m, p, plot=self.doplot )
    """
    def testExpModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******EXP**************************" )
        m = ExpModel( )
        p = numpy.asarray( [1.2,-0.1], dtype=float )

        stdModeltest( m, p, plot=self.doplot )

        print( "******DECAY************************" )

        m = ExpModel( decay=True )
        p = numpy.asarray( [1.2,0.1], dtype=float )

        stdModeltest( m, p, plot=self.doplot )

    def testLogisticModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******LOGISTIC***********************" )
        m = LogisticModel( )
        p = numpy.asarray( [1.2,-0.2,0.3], dtype=float )

        stdModeltest( m, p, plot=self.doplot )

    def testLorentzModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******LORENTZ***********************" )
        m = LorentzModel( )
        p = numpy.asarray( [1.2,-0.2,0.3], dtype=float )

        stdModeltest( m, p, plot=self.doplot )

    def testVoigtModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******VOIGT***********************" )
        m = VoigtModel( )
        p = numpy.asarray( [1.2,0.2,0.3,0.4], dtype=float )
        print( p )

        stdModeltest( m, p, plot=self.doplot ) #, warn=["nopart"] )

    def testPseudoVoigtModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******PSEUDOVOIGT***********************" )
        m = PseudoVoigtModel( )
        p = numpy.asarray( [1.2,0.2,0.3,0.4], dtype=float )
        print( p )

        stdModeltest( m, p, plot=self.doplot )

    def testFreeShapeModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******FREESHAPE********************" )
        m = FreeShapeModel( 5, pixperbin=1, xlo=-1, xhi=1.1, nconvolve=2 )
        self.assertTrue( m.npbase == 5 )
        p = numpy.asarray( [1.2,0.2,1.0,0.5,0.3], dtype=float )

        stdModeltest( m, p, plot=self.doplot )

    def testGaussModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******GAUSS*************************" )
        m = GaussModel( )
        p = numpy.asarray( [1.2,-0.2,0.3], dtype=float )

        stdModeltest( m, p, plot=self.doplot )

    def testGaussPlusBackgroundModel( self ):
        print( "******GAUSS + BG**********************" )
        gm = GaussModel( )
        print( gm )
        print( gm.parameters )
        pm = PolynomialModel( 1 )
        print( pm )
        print( pm.parameters )
        gm.addModel( pm )
        par = numpy.asarray( [3,0.2,0.2,0.1,0.1], dtype=float )

        stdModeltest( gm, par, plot=self.doplot )

    def testHarmonicModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******HARMONIC**********************" )
        m = HarmonicModel( 1 )
        self.assertTrue( m.order == 1 )
        self.assertTrue( m.npbase == 2 )
        self.assertTrue( m.period == 1.0 )

        m = HarmonicModel( order=3, period=0.4 )
        self.assertTrue( m.order == 3 )
        self.assertTrue( m.npbase == 6 )
        self.assertTrue( m.period == 0.4 )

        p = numpy.asarray( [1.2,-0.2,0.3,1.2,-0.2,0.3], dtype=float )

        stdModeltest( m, p, plot=self.doplot )

    def testPolynomialModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******POLYNOMIAL**********************" )
        m = PolynomialModel( 3 )
        self.assertTrue( m.getNumberOfParameters( ) == 4 )
        self.assertTrue( m.npbase == 4 )
        p = numpy.asarray( [1,-2,3,-2], dtype=float )

        stdModeltest( m, p, plot=self.doplot )

    def testFixedPolynomialModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******POLYNOMIAL FIXED****************" )
        m = PolynomialModel( 5, fixed={1:0.4, 3:-1.2} )
        self.assertTrue( m.getNumberOfParameters( ) == 4 )
        self.assertTrue( m.npbase == 4 )
        self.assertTrue( m.npmax == 6 )
        p = numpy.asarray( [1,-2,3,-2], dtype=float )

        stdModeltest( m, p, plot=self.doplot )

    def testChebyshevPolynomialModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******CHEBYSHEV POLYNOMIAL*************" )
        m = ChebyshevPolynomialModel( 4 )
        self.assertTrue( m.getNumberOfParameters( ) == 5 )
        self.assertTrue( m.npbase == 5 )
        p = numpy.asarray( [1,-2,3,-2,0.3], dtype=float )

        stdModeltest( m, p, plot=self.doplot )

    def testPadeModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******PADE**********************" )
        m = PadeModel( 3, 2 )
        self.assertTrue( m.getNumberOfParameters( ) == 6 )
        self.assertTrue( m.npbase == 6 )
        p =  numpy.asarray( [1,-2,3,-2,0.3,1], dtype=float )
        stdModeltest( m, p, plot=self.doplot )

    def testPowerLawModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******POWERLAW*******************" )
        m = PowerLawModel( )
        self.assertTrue( m.npchain == 3 )
        self.assertTrue( m.npbase == 3 )
        p = [2.3, -1.1, 0.5]
        stdModeltest( m, p, plot=self.doplot )

    def testPowerModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******POWER**********************" )
        m = PowerModel( 3 )
        self.assertTrue( m.getNumberOfParameters( ) == 1 )
        self.assertTrue( m.npbase == 1 )
        p = [2.3]
        stdModeltest( m, p, plot=self.doplot )

    def testRadialVelocityModel( self ):
        x  = numpy.linspace( 0, 1400, 1401, dtype=float )
        print( "******RADIALVELOCITY***************" )
        m = RadialVelocityModel( )
        self.assertTrue( m.npars == 5 )
        self.assertTrue( m.npbase == 5 )
        p = [0.67, 130, 1200.0, 4.0, 3.0]
        if self.doplot :
            for p3 in range( 7 ) :
                p[4] = 0.7 + 0.5 * p3
                y = m.result( x, p )
                plt.plot( x, y + 100 * p3, '-' )
#                yo = m.baseResultXXX( x, p )
#                plt.plot( x, yo + 100 * p3, '.' )

            plt.show()

        stdModeltest( m, p, plot=self.doplot )

    def testRadialVelocityModel_2( self ):
        x  = numpy.linspace( 0, 1400, 1401, dtype=float )
        print( "******RADIALVELOCITY***************" )
        m = RadialVelocityModel( spherical=False )
        self.assertTrue( m.npars == 5 )
        self.assertTrue( m.npbase == 5 )
        p = [0.67, 130, 1200.0, 4.0, 3.0]
        if self.doplot :
            for p3 in range( 7 ) :
                p[4] = 0.7 + 0.5 * p3
                y = m.result( x, p )
                plt.plot( x, y + 100 * p3, '-' )
#                yo = m.baseResultXXX( x, p )
#                plt.plot( x, yo + 100 * p3, '.' )

            plt.show()

        stdModeltest( m, p, plot=self.doplot )

    def testSincModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SINC***********************" )
        m = SincModel( )
        self.assertTrue( m.npchain == 3 )
        self.assertTrue( m.npbase == 3 )
        p = [2.3, -1.1, 0.5]
        stdModeltest( m, p, plot=self.doplot )

    def testSineModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SINE***********************" )
        m = SineModel( )
        self.assertTrue( m.npchain == 3 )
        self.assertTrue( m.npbase == 3 )
        p = [1.3, -1.1, 0.5]
        stdModeltest( m, p, plot=self.doplot )

    def testPhaseSineModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SINE***********************" )
        m = SineModel( phase=True )
        self.assertTrue( m.npchain == 3 )
        self.assertTrue( m.npbase == 3 )
        p = [1.3, -1.1, 0.5]
        stdModeltest( m, p, plot=self.doplot )

    def testSineAmpModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SINEAMP********************" )
        m = SineAmpModel( 1.3 )
        self.assertTrue( m.npchain == 2 )
        self.assertTrue( m.npbase == 2 )
        p = [-1.1, 0.5]
        stdModeltest( m, p, plot=self.doplot )

    def testSineSplineModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SINESPLINE******************" )
        knots = numpy.arange( 3, dtype=float ) - 1
        m = SineSplineModel( 1.3, knots )
        self.assertTrue( m.npchain == 10 )
        self.assertTrue( m.npbase == 10 )
        p = [0.0, 0.5, -0.1, 0.5, -0.1, 0.0, +1.1, 0.5, -1.1, -0.5]
        amp = m.getAmplitudes( x, p )
        print( amp[0], amp[1] )
        stdModeltest( m, p, plot=self.doplot )

    def testSineSplineDriftModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SINESPLINEDRIFT*************" )
        knots = numpy.arange( 3, dtype=float ) - 1
        m = SineSplineDriftModel( knots, degree=1 )
        print( m.npchain )
        self.assertTrue( m.npchain == 12 )
        self.assertTrue( m.npbase == 12 )
        p = [0.0, 0.1, 0.0, 0.5, -0.1, 0.5, -0.1, 0.0, +1.1, 0.5, -1.1, -0.5]
        amp = m.getAmplitudes( x, p )
        print( amp[0], amp[1] )
        stdModeltest( m, p, plot=self.doplot )

    def testSineDriftModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SINEDRIFT******************" )
        m = SineDriftModel( )
        self.assertTrue( m.npchain == 4 )
        self.assertTrue( m.npbase == 4 )
        p = [1.1, 0.5, 0.8, 0.4 ]
        stdModeltest( m, p, plot=self.doplot )

    def testBSplinesModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******B-SPLINES*************************" )
        knots = numpy.arange( 11, dtype=float )
        m0 = BSplinesModel( knots=knots )
        self.assertTrue( m0.order == 3 )
        self.assertTrue( m0.getNumberOfParameters( ) == 13 )
        self.assertTrue( m0.npbase == 13 )
        m1 = BSplinesModel( nrknots=5, min=-1, max=1 )
        self.assertTrue( m1.order == 3 )
        self.assertTrue( m1.getNumberOfParameters( ) == 7 )
        self.assertTrue( m1.npbase == 7 )
        for k in range( len( m1.knots ) ) :
            self.assertTrue( m1.knots[k] == -1 + k / 2 )
        xin = numpy.linspace( 0, 10, 101, dtype = float )
        m2 = BSplinesModel( xrange=xin, order=2, nrknots=11 )
        self.assertTrue( m2.order == 2 )
        self.assertTrue( m2.getNumberOfParameters( ) == 12 )
        self.assertTrue( m2.npbase == 12 )
        for k in range( len( m2.knots ) ) : self.assertTrue( m2.knots[k] == k )

        self.assertRaises( ValueError, BSplinesModel )

        m1 = BSplinesModel( nrknots=5, order=3, min=-1.01, max=1.01 )
        p = numpy.asarray( [2,1,2,1,2,1,2], dtype=float )
        p = p[:m1.npchain]
        print( m1.npchain, p )
        stdModeltest( m1, p, plot=self.doplot )

    def testBasicSplinesModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******BASICSPLINES*************************" )
        knots = numpy.arange( 11, dtype=float )
        m0 = BasicSplinesModel( knots=knots )
        self.assertTrue( m0.order == 3 )
        self.assertTrue( m0.getNumberOfParameters( ) == 13 )
        self.assertTrue( m0.npbase == 13 )

        m1 = BasicSplinesModel( nrknots=5, min=-1, max=1 )
        self.assertTrue( m1.order == 3 )
        self.assertTrue( m1.getNumberOfParameters( ) == 7 )
        self.assertTrue( m1.npbase == 7 )
        for k in range( len( m1.knots ) ) :
            self.assertTrue( m1.knots[k] == -1 + k / 2 )

        xin = numpy.linspace( 0, 10, 101, dtype = float )
        m2 = BasicSplinesModel( xrange=xin, order=2, nrknots=11 )
        self.assertTrue( m2.order == 2 )
        self.assertTrue( m2.getNumberOfParameters( ) == 12 )
        self.assertTrue( m2.npbase == 12 )
        for k in range( len( m2.knots ) ) :
            self.assertTrue( m2.knots[k] == k )

        self.assertRaises( ValueError, BasicSplinesModel )

        m1 = BasicSplinesModel( nrknots=5, order=3, min=-1.01, max=1.01 )
        p = numpy.asarray( [2,1,2,1,2,1,2], dtype=float )
        print( m1.npchain, p )
        stdModeltest( m1, p, plot=self.doplot )

    def testSplinesModel( self ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        print( "******SPLINES*************************" )
        knots = numpy.arange( 11, dtype=float )
        m0 = SplinesModel( knots=knots )
        self.assertTrue( m0.order == 3 )
        self.assertTrue( m0.getNumberOfParameters( ) == 13 )
        self.assertTrue( m0.npbase == 13 )

        m1 = SplinesModel( nrknots=5, min=-1, max=1 )
        self.assertTrue( m1.order == 3 )
        self.assertTrue( m1.getNumberOfParameters( ) == 7 )
        self.assertTrue( m1.npbase == 7 )
        for k in range( len( m1.knots ) ) :
            self.assertTrue( m1.knots[k] == -1 + k / 2 )

        xin = numpy.linspace( 0, 10, 101, dtype = float )
        m2 = SplinesModel( xrange=xin, order=2, nrknots=11 )
        self.assertTrue( m2.order == 2 )
        self.assertTrue( m2.getNumberOfParameters( ) == 12 )
        self.assertTrue( m2.npbase == 12 )
        for k in range( len( m2.knots ) ) : self.assertTrue( m2.knots[k] == k )

        self.assertRaises( ValueError, SplinesModel )

        p = numpy.asarray( [0,1,0,1,0,1,0], dtype=float )
        stdModeltest( m1, p, plot=self.doplot )



    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestModels.__class__ )

if __name__ == '__main__':
    unittest.main( )





import numpy as numpy
import math
from . import Tools
from .Tools import setAttribute as setatt
from .NonLinearModel import NonLinearModel
from .LorentzModel import LorentzModel
from .GaussModel import GaussModel

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2016 - 2020 Do Kester

class PseudoVoigtModel( NonLinearModel ):
    """
    Approximation of VoigtModel as the sum of a GaussModel and a LorentzModel

        F(x:p) = p_3 * L(x:p) + ( 1 - p_3 ) * G(x:p)

    where L() and G() are the LorentzModel and the GaussModel, resp. and p_3
    is the fractional contribution of them. 0 < p_3 < 1.

    The models takes 4 parameters: amplitude, center frequency, half-width and
    the balance between the models
.
    These are initialised to [1, 0, 1, 0.5].
    Parameter 2 (width) is always kept positive ( >=0 ).

    Examples
    --------
    >>> voigt = PseudoVoigtModel( )
    >>> voigt.setParameters( [5, 4, 1, 0.7] )
    >>> print( voigt( numpy.arange(  41 , dtype=float ) / 5 ) )      # from [0,8]

    Attributes
    ----------
    gauss : GaussModel
        to construct the gauss parts
    lorentz : LorentzModel
        to construct the lorentz parts

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames

    """

    def __init__( self, copy=None, **kwargs ):
        """
        PseudoVoigt model.
        <br>
        Number of parameters is 4.

        Parameters
        ----------
        copy : PseudoVoigtModel
            to be copied
        fixed : None or dictionary of {int:float|Model}
            int         index of parameter to fix permanently.
            float|Model values for the fixed parameters.
            Attribute fixed can only be set in the constructor.
            See: @FixedModel

        """
        param = [1.0, 0.0, 1.0, 0.5]
        names = ["amplitude","center","width","balance"]

        super( PseudoVoigtModel, self ).__init__( 4, copy=copy, params=param,
                        names=names, **kwargs )

        setatt( self, "gauss", GaussModel() )
        setatt( self, "lorentz", LorentzModel() )
        if copy is None :
            self.posIndex = [2]

    def copy( self ):
        """ Copy method.  """
        return PseudoVoigtModel( copy=self )

    def __setattr__( self, name, value ):
        """
        Set attributes.

        Parameters
        ----------
        name :  string
            name of the attribute
        value :
            value of the attribute

        """
        if name == "gauss" :
            setatt( self, name, value, type=GaussModel )
        elif name == "lorentz" :
            setatt( self, name, value, type=LorentzModel )
        else :
            super( PseudoVoigtModel, self ).__setattr__( name, value )


    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Note:
        1. The "balance" parameter (item 3) should be kept between [0..1]
        2. The "width" parameter (item 2)
        the width in the parameter array ( items 2 & 3 ) are kept
        strictly positive. I.e. they are changed when upon xdata they are negative.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        return ( params[3] * self.lorentz.baseResult( xdata, params[:3] ) +
                 ( 1 - params[3] ) * self.gauss.baseResult( xdata, params[:3] ) )

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the xdata value.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            values for the parameters.
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        pl = parlist
        if parlist is not None and 3 in pl :
            pl = parlist.copy()
            pl.remove( 3 )

        pars = params[:3]
        partial = ( params[3] * self.lorentz.basePartial( xdata, pars, parlist=pl ) +
                 ( 1 - params[3] ) * self.gauss.basePartial( xdata, pars, parlist=pl ) )

        if parlist is None or 3 in parlist :
            part3 = self.lorentz.baseResult( xdata, pars ) - self.gauss.baseResult( xdata, pars )
            (n0,n1) = partial.shape
            shp = ( n0, n1+1 )
            pp = numpy.zeros( shp, dtype=float )
            pp[:,:n1] = partial
            pp[:,n1] = part3
            partial = pp

        return partial

    def baseDerivative( self, xdata, params ) :
        """
        Return the derivative df/dx at each xdata (=x).

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the derivative
        params : array_like
            values for the parameters.

        """
        return ( params[3] * self.lorentz.baseDerivative( xdata, params[:3] ) +
                 ( 1 - params[3] ) * self.gauss.baseDerivative( xdata, params[:3] ) )


    def baseName( self ):
        """
        Returns a string representation of the model.
        """
        return str( "PseudoVoigt: p_3 * Lorentz + ( 1 - p_3 ) * Gauss" )

    def baseParameterUnit( self, k ):
        """
        Return the name of a parameter.

        Parameters
        ---------
        k : int
            parameter number.

        """
        if k == 0:
            return self.yUnit
        return self.xUnit





# run with : python3 -m unittest TestKernels

import unittest
import os
import numpy as numpy
from astropy import units
import matplotlib.pyplot as plt
import warnings

from BayesicFitting import *
from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestKernels( unittest.TestCase ):
    """
    Test harness for Models

    Author:      Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )


    def testHuber( self ) :
        kernel = Huber()
        xhm = kernel.fwhm / 2
        self.assertAlmostEqual( kernel.result( xhm ), 0.5 )
        self.assertAlmostEqual( kernel.result( -xhm ), 0.5 )
        if self.doplot :
            self.plotK( [Huber] )

    def testGauss( self ) :
        print( "**** Gauss ********************" )
        km = KernelModel( kernel=Gauss() )
        gm = GaussModel()
        k2 = Kernel2dModel( kernel=Gauss() )

        x  = numpy.asarray( [-1.0, -0.4, 0.0, 0.4, 1.0] )
        x += 0.02
        p = [1.0, 0.0, 0.2]
        x2 = numpy.append( x, x ).reshape( 2, -1 ).T
        p2 = [1.0, 0.0, 0.0, 0.2]
        print( fmt( x2 ) )

        print( fmt( gm.result( x, p ) ) )
        print( fmt( km.result( x, p ) ) )
        print( fmt( k2.result( x2, p2 ) ) )
        print( fmt( gm.partial( x, p ) ) )
        print( fmt( km.partial( x, p ) ) )
        print( fmt( k2.partial( x2, p2 ) ) )
        print( fmt( gm.derivative( x, p ) ) )
        print( fmt( km.derivative( x, p ) ) )
        print( fmt( k2.derivative( x2, p2 ) ) )
        
        x3 = numpy.asarray( [[-1.0, -0.8], [-0.6, -0.4], [-0.2, 0.0], [0.2, 0.4], [0.6, 0.8],
                [1.0, -1.0], [-0.8, -0.6], [-0.4, -0.2], [0.0, 0.2], [0.4, 0.6], [0.8, 1.0]] )
        p3 = [-1.1, 0.5, 0.04, 1.2]

        print( "k2 df  ", fmt( k2.derivative( x3[0:1,:], p3 ), max=None ) )
        print( "k2 num ", fmt( k2.strictNumericDerivative( x3[0:1,:], p3 ), max=None ) )
        print( fmt( gm.derivative( x3[0:1,0], [-1.1, 0.5, 1.2] ) ) )
        print( fmt( gm.strictNumericDerivative( x3[0:1,0], [-1.1, 0.5, 1.2] ) ) )
        print( fmt( gm.derivative( x3[0:1,1], [-1.1, 0.04, 1.2] ) ) )
        print( fmt( gm.strictNumericDerivative( x3[0:1,1], [-1.1, 0.04, 1.2] ) ) )


    def testKernels( self ):
        kernels = [Gauss, Lorentz, Sinc, Biweight, Cosine, CosSquare, Parabola, Triangle,
                   Tricube, Triweight, Uniform]
        if self.doplot :
            self.plotK( kernels )

        for kernl in kernels :
            kernel = kernl()
            self.stdKerneltest( kernel, plot=self.doplot )
            self.assertTrue( kernel.isBound() or ( kernl in kernels[:3] ) )


    def testTHC( self ) :
        x = numpy.linspace( -5, 5, 1001 )
        for kc in range( 7 ) :

            kernel = Tophat( nconv=kc )
            print( "****"  , kernel, "  *************************" )

            y = kernel.result( x )
            top = y[500]

            kernin = numpy.sum( y ) / 100
            self.assertAlmostEqual( kernin, kernel.integral )

            print( kc, top, numpy.sum( y )/100 )
            if self.doplot :
                plt.plot( x, kernel.result( x ) )
                plt.plot( x, kernel.partial( x ) )

            self.stdKerneltest( kernel )

        if self.doplot :
            plt.show()

    def testonex( self ) :
        sc = Sinc()
        ysc = sc.result( 0.0 )
        print( "sinc  ", ysc, ysc.__class__ )

        un = Uniform()
        yun = un.result( 0.0 )
        print( "unif  ", yun, yun.__class__ )

        for n in range( 7 ) :
            thc = Tophat( nconv=n )
            ytc = thc.result( 0.0 )
            print( "thc%d  " % n, ytc, ytc.__class__ )



    def stdKerneltest( self, kernel, plot=None ):
        x  = numpy.asarray( [-1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0] )
        x += 0.02

        model = KernelModel( kernel=kernel )
        print( "********************************************" )
        print( model )

        numpy.set_printoptions( precision=3, suppress=False )

        par = [1.0,0.0,1.0]
        print( model.partial( x, par ) )
        model.testPartial( x[5], model.parameters )

        model.testPartial( x[0], par )
        model.xUnit = units.m
        model.yUnit = units.kg
        for k in range( model.getNumberOfParameters() ):
            print( "%d  %-12s  %-12s"%(k, model.getParameterName( k ), model.getParameterUnit( k ) ) )

        xx = numpy.linspace( 0, 101, 101000 )
        yy = model.result( xx )
        sn = numpy.sum( yy ) / 500
        ss = kernel.integral

        print( "Integral   ", sn, model.getIntegralUnit( ) )
        self.assertAlmostEqual( sn, 1.0, 1 )
        self.assertAlmostEqual( ss, 1.0 / model.parameters[0], 4 )

        xhm = kernel.fwhm / 2
        yhm = model.result( xhm )[0]
        self.assertAlmostEqual( yhm, model.result( -xhm ) )
        y0 = model.result( 0.0 )[0]
        print( xhm, yhm, y0, yhm / y0, yhm.__class__ )

        self.assertAlmostEqual( (yhm / y0), 0.5 )

        part = model.partial( x, par )
        nump = model.numPartial( x, par )
        print( part.shape, nump.shape )
        for k in range( 11 ) :
            print( "%3d %8.3f"%(k, x[k]), part[k,:], nump[k,:] )

        k = 0
        np = model.npbase
        for (pp,nn) in zip( part.flatten(), nump.flatten() ) :
            print( "%3d %10.4f  %10.4f %10.4f"%(k, x[k//np], pp, nn) )
            self.assertAlmostEqual( pp, nn, 3 )
            k += 1

        mc = model.copy( )
        for k in range( mc.getNumberOfParameters() ):
            print( "%d  %-12s  %-12s"%(k, mc.getParameterName( k ), mc.getParameterUnit( k ) ) )
        mc.parameters = par
        model.parameters = par

        for (k,xk,r1,r2) in zip( range( x.size ), x, model.result( x ), mc.result( x ) ) :
            print( "%3d %8.3f  %10.3f %10.3f" % ( k, xk, r1, r2 ) )
            self.assertEqual( r1, r2 )

    def plotK( self, kernels ) :
        x = numpy.linspace( -5, +5, 1001 )
        par = [1.0,0.0,1.0]
        for kernel in kernels :
            model = kernel()
            print( model )
            y = model.result( x )
            plt.plot( x, y, '-', linewidth=2 )
            x2 = numpy.linspace( -4, +4, 11 )
            dy = model.partial( x2 )
            print( dy )
            yy = model.result( x2 )
            print( yy )
            for k in range( 11 ) :
                x3 = numpy.asarray( [-0.05, +0.05] )
                y3 = x3 * dy[k] + yy[k]
                plt.plot( x2[k] + x3, y3, 'r-' )

        plt.show()


    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestModels.__class__ )

if __name__ == '__main__':
    unittest.main( )





# run with : python3 -m unittest TestScale

import unittest
import os
import numpy as numpy
from astropy import units
import math
from numpy.testing import assert_array_almost_equal as assertAAE


from BayesicFitting import *
from BayesicFitting import formatter as fmt

import matplotlib.pyplot as plt

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class TestWeight2( unittest.TestCase ):
    """
    Test harness for Engines

    Author       Do Kester

    """
    def __init__( self, name ):
        super( TestWeight2, self ).__init__( name )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )


    def test1( self ):

        numpy.random.seed( 345347 )

        bs = 100
        bof = 6 * bs
        xof = 0.5 / bs
        nbin = 2 * bof + 1

        for np in [10, 100, 1000, 10000, 100000] :
            mdl1 = PolynomialModel( 0 )
            x = numpy.random.randn( np )

            ftr1 = Fitter( x, mdl1 )
            p1 = ftr1.fit( x )[0]
            s1 = ftr1.scale
            t1 = ftr1.stdevs[0]
            r1 = ftr1.stdevScale
            c1 = ftr1.chisq
            h1 = ftr1.hessian
            v1 = ftr1.covariance
            e1a = ftr1.getEvidence( limits=[-10,10], noiseLimits=[0.1,10] )
            o1a = ftr1.logOccam
            l1a = ftr1.logLikelihood
            e1b = ftr1.getEvidence( limits=[-10,10] )
            o1b = ftr1.logOccam
            l1b = ftr1.logLikelihood

            yy = numpy.zeros( nbin )
            xs = numpy.array( bs * x + bof, dtype=int )
            unq, cnt = numpy.unique( xs, return_counts=True )
            yy[unq] += cnt
            xx = numpy.linspace( -6, +6, nbin )
#            print( x )
#            print( unq )
#            print( xx[unq] )
#            print( cnt )
            mdl2 = PolynomialModel( 0 )
            ftr2 = Fitter( xx, mdl2 )
            p2 = ftr2.fit( xx, weights=yy )[0]

            s2 = ftr2.scale
            t2 = ftr2.stdevs[0]
            r2 = ftr2.stdevScale
            c2 = ftr2.chisq
            h2 = ftr2.hessian
            v2 = ftr2.covariance
            e2a = ftr2.getEvidence( limits=[-10,10], noiseLimits=[0.1,10] )
            o2a = ftr2.logOccam
            l2a = ftr2.logLikelihood
            e2b = ftr2.getEvidence( limits=[-10,10] )
            o2b = ftr2.logOccam
            l2b = ftr2.logLikelihood

            print( "Points  ", fmt( np ) )
            print( "Params  ", fmt( p1 ), " +- ", fmt( t1 ) )
            print( "Params  ", fmt( p2 ), " +- ", fmt( t2 ) )
            print( "Scale   ", fmt( s1 ), " +- ", fmt( r1 ) )
            print( "Scale   ", fmt( s2 ), " +- ", fmt( r2 ) )
            print( "Chisq   ", fmt( c1 ) )
            print( "Chisq   ", fmt( c2 ) )
            print( "Hessian ", fmt( h1 ) )
            print( "Hessian ", fmt( h2 ) )
            print( "Covar   ", fmt( v1 ) )
            print( "Covar   ", fmt( v2 ) )
            print( "Evid 1a ", fmt(e1a ), fmt( o1a ), fmt( l1a ) )
            print( "Evid 1b ", fmt(e1b ), fmt( o1b ), fmt( l1b ) )
            print( "Evid 2a ", fmt(e2a ), fmt( o2a ), fmt( l2a ) )
            print( "Evid 2b ", fmt(e2b ), fmt( o2b ), fmt( l2b ) )
            print( "         " )

            self.assertAlmostEqual( p1, p2, 1 )
            self.assertAlmostEqual( t1, t2, 2 )
            self.assertAlmostEqual( s1, s2, 1 )
            self.assertAlmostEqual( r1, r2, 2 )
#            self.assertAlmostEqual( c1, c2, 0 )
            self.assertAlmostEqual( h1, h2, 5 )
            assertAAE( v1, v2, 2 )
            self.assertAlmostEqual( e1a, e2a, 0 )
            self.assertAlmostEqual( e1b, e2b, 0 )
            self.assertAlmostEqual( o1a, o2a, 1 )
            self.assertAlmostEqual( o1b, o2b, 1 )
            self.assertAlmostEqual( l1a, l2a, 0 )
            self.assertAlmostEqual( l1b, l2b, 0 )


    def test2( self ) :

        numpy.random.seed( 345347 )

        bs = 100
        bof = 6 * bs
        xof = 0.5 / bs
        nbin = 2 * bof + 1

        np = 1000
        x0 = numpy.linspace( -9.995, 9.995, np )
        y0 = 0.1 * x0 + 0.3 * numpy.random.randn( np )

        if self.doplot : plt.plot( x0, y0, 'k.' )

        x1 = []
        y1 = []
        w1 = []
        x2 = []
        y2 = []
        w2 = []
        for kx in range( -10, 10 ) :
            qx = numpy.where( numpy.logical_and( x0 >= kx, x0 < (kx+1) ) )
            xq = x0[qx]
            yq = y0[qx]
            for ky in range( 40 ) :
                yk = -2.0 + 0.1 * ky
                q = numpy.where( numpy.logical_and( yq >= yk, yq < (yk+0.1) ) )[0]
                lenq = len( q )
                if lenq == 0: continue
                x1 = x1 + [kx + 0.5]
                y1 = y1 + [yk + 0.05]
                w1 = w1 + [len( q )]

        x2 = numpy.asarray( ( x0 + 20 ), dtype=int ) - 19.5
        y2 = numpy.asarray( ( 10 * y0 + 40 ), dtype=int ) / 10 - 3.95


        print( len(x1) )
        print( "x0 ", fmt( x0, max=8, tail=4 ) )
        print( "x2 ", fmt( x2, max=8, tail=4 ) )
        print( "y0 ", fmt( y0, max=8, tail=4 ) )
        print( "y2 ", fmt( y2, max=8, tail=4 ) )
        print( "x1 ", fmt( x1, max=8, tail=4 ) )
        print( "y1 ", fmt( y1, max=8, tail=4 ) )
        print( "w1 ", fmt( w1, max=8, tail=4 ) )
#        print( fmt( w2, max=10 ) )

        md0 = PolynomialModel( 1 )
        ft0 = Fitter( x0, md0 )
        (np0,par0,std0,scl0,sds0,chi0,hes0,cov0,evi0,occ0,lik0) = self.stdfit( md0, ft0, y0 )

        md1 = PolynomialModel( 1 )
        ft1 = Fitter( x1, md1 )
        (np1,par1,std1,scl1,sds1,chi1,hes1,cov1,evi1,occ1,lik1) = self.stdfit( md1, ft1, y1, w=w1 )

        md2 = PolynomialModel( 1 )
        ft2 = Fitter( x2, md2 )
        (np2,par2,std2,scl2,sds2,chi2,hes2,cov2,evi2,occ2,lik2) = self.stdfit( md2, ft2, y2 )

        self.assertTrue( np0 == np1 == np2 )

        assertAAE( par0, par1, 2 )
        assertAAE( par2, par1 )
        assertAAE( std0, std1, 2 )
        assertAAE( std2, std1 )
        self.assertAlmostEqual( scl0, scl1, 2 )
        self.assertAlmostEqual( scl2, scl1 )
        self.assertAlmostEqual( sds0, sds1, 2 )
        self.assertAlmostEqual( sds2, sds1 )
        self.assertAlmostEqual( chi2, chi1 )
        assertAAE( hes2, hes1 )
        self.assertAlmostEqual( evi2, evi1 )
        self.assertAlmostEqual( occ2, occ1 )
        self.assertAlmostEqual( lik2, lik1 )

        if self.doplot :
            plt.plot( x1, y1, 'r*' )
            plt.plot( x2, y2, 'g+' )
            plt.show()



    def stdfit( self, mdl, ftr, y, w=None ) :

        np = len( y ) if w is None else numpy.sum( w )
        par = ftr.fit( y, weights=w )
        std = ftr.stdevs
        scl = ftr.scale
        sds = ftr.stdevScale
        chi = ftr.chisq
        hes = ftr.hessian
        cov = ftr.covariance
        evi = ftr.getEvidence( limits=[-10,10], noiseLimits=[0.1,10] )
        occ = ftr.logOccam
        lik = ftr.logLikelihood

        print( "Points  ", fmt( np ) )
        print( "Params  ", fmt( par ), " +- ", fmt( std ) )
        print( "Scale   ", fmt( scl ), " +- ", fmt( sds ) )
        print( "Chisq   ", fmt( chi ) )
        print( "Hessian ", fmt( hes, indent=9 ) )
        print( "Covar   ", fmt( cov, indent=9 ) )
        print( "Evid    ", fmt( evi ), fmt( occ ), fmt( lik ) )

        return (np,par,std,scl,sds,chi,hes,cov,evi,occ,lik)

    def suite( cls ):
        return unittest.TestCase.suite( TestWeight2.__class__ )


if __name__ == '__main__':
    unittest.main( )





import numpy as numpy
from .Model import Model

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "2.8.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (JAVA code)
#  *    2016 - 2021 Do Kester

class LinearModel( Model ):
    """
    Anchestor of all linear models.

    LinearModel implements the baseResult method needed in all linear models.

    For Linear models it holds that

        f( x:p ) = &sum;( p_i * df( x )/dp_i )

    which means that only the partial derivatives to p_i need to be given
    as basePartial. The baseResult follows directly from that one.
    It is implemented here.

    Attributes
    ----------
    None of its own

    Attributes from Model
    ---------------------
        parameters, stdevs, npchain
        _next, _head, _operation
        xUnit, yUnit (relegated to model)

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames

    Author       Do Kester


    """
    def __init__( self, nparams, ndim=1, copy=None, **kwargs ):
        """
        class for all linear models.

        Parameters
        ----------
        nparams : int
            the number of parameters in this model
        ndim : int
            the dimensionality of the inputs (default: 1)
        copy : LinearModel
            model to be copied (default: None)
        kwargs : dict
            Possibly includes keywords from
                @Model :        params
                @FixedModel :   fixed, names
                @BaseModel :    posIndex, nonZero

        """
        super( LinearModel, self ).__init__( nparams, ndim=ndim, copy=copy, **kwargs )

    def baseResult( self, xdata, params ):
        """
        Returns the base result of linear models.

        for linear models the result is the inner product of parameters
        and partial derivatives.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        part = self.basePartial( xdata, params )

        res = numpy.zeros( part.shape[0], dtype=float )
        ## params might be a heterogeneous list of float or [float] from FixedModel
        ## numpy.inner does not work here
        for k in range( self.npmax ) :
            res += params[k] * part[:,k]

        return res



import numpy as numpy
import math
import warnings
from . import Tools

from .Engine import Engine
from .Engine import DummyPlotter
from .OrthonormalBasis import OrthonormalBasis
from .Formatter import formatter as fmt
from .Formatter import fma

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "2.7.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2010 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2021 Do Kester

class ChordEngine( Engine ):
    """
    Move a a walker in a random direction.

    The ChordEngine draws a random line through the walker parameters in
    unit space, from unitMin (lowpoint) with lengths unitRange (highpoint).

    A random point on the line is selected. If the corresponding parameter
    set has a likelihood < LowLhood, it is accepted. Otherwise either the
    highpoint is reset to the random point (if randompoint > walkerpoint)
    or the lowpoint is replaced by the randompoint (if walker < random).
    Then a new random point on the line is selected, until the point is accepted.

    When the point is accepted, another random line is constructed
    through the new point and orthogonal to (all) previous ones.
    (The orthogonality is not implemented now. TBC).

    This is an independent implementation inspired by the polychord engine
    described in:
    "POLYCHORD: next-generation nested sampling",
    WJ Handley, MP Hobson and AN Lasenby.
    MNRAS (2015) Volume 453, Issue 4, p 4384–4398

    Attributes
    ----------
    nstep : int
        average number of orthogonal steps
    debug : bool
        perform the step-out action too

    Attributes from Engine
    ----------------------
    walkers, errdis, slow, maxtrials, rng, verbose, report, unitRange, unitMin

    Author       Do Kester.

    """
    #  *********CONSTRUCTORS***************************************************
    def __init__( self, walkers, errdis, copy=None, **kwargs ) :
        """
        Constructor.

        Parameters
        ----------
        walkers : WalkerList
            walkers to be diffused
        errdis : ErrorDistribution
            error distribution to be used
        copy : ChordEngine
            to be copied
        kwargs : for Engine
            "slow", "seed", "verbose"
        """
        super( ).__init__( walkers, errdis, copy=copy, **kwargs )

        self.nstep = 5
        self.maxtrials = 25
        self.plotter = DummyPlotter()
        self.debug = False

    def copy( self ):
        """ Return copy of this.  """
        return ChordEngine( self.walkers, self.errdis, copy=self )

    def __str__( self ):
        return str( "ChordEngine" )


    #  *********EXECUTE***************************************************
    def execute( self, kw, lowLhood, append=False, iteration=0 ):
        """
        Execute the engine by diffusing the parameters.

        Parameters
        ----------
        kw : int
            index of walker to diffuse
        lowLhood : float
            lower limit in logLikelihood
        append : bool
            set walker in place or append
        iteration : int
            iteration number

        Returns
        -------
        int : the number of successfull moves

        """
        self.reportCall()

        walker = self.walkers[kw].copy()
        problem = walker.problem
        fitIndex = walker.fitIndex
        np = len( fitIndex )

        param = walker.allpars
        usav = self.domain2Unit( problem, param[fitIndex], kpar=fitIndex )
        self.plotter.start( param=param )

        # Determine n-dim box boundaries in which the parameters are located
        if np > len( self.unitRange ) :
            self.unitRange = numpy.ones( np, dtype=float )
            self.unitMin = numpy.zeros( np, dtype=float )

        uran = self.unitRange[fitIndex]
        umin = self.unitMin[fitIndex]

        dur = 2 * numpy.max( uran ) / len( self.walkers )       # extend range a bit
        uran += 2 * dur
        umin = self.unitMin[fitIndex] - dur
        umax = umin + uran

        # make sure that 0 <= umin < usav < umax <= 1
        umin = numpy.where( usav < umin, 2 * usav - umin, umin )
        umax = numpy.where( usav > umax, 2 * usav - umax, umax )
        umin = numpy.maximum( umin, 0 )
        umax = numpy.minimum( umax, 1 )

        # number of cycles in ChordEngine
        nstep = int( self.nstep * ( 1 + self.rng.rand() ) )

        onb = OrthonormalBasis( )
        # random direction in n-dim unit-parameter space
        vel = self.rng.rand( np ) - 0.5

        if self.verbose > 4 :
            print( "++++++++++++++++++++++++++++++++++++++++++" )
            print( walker.id, nstep, fmt(lowLhood), fmt( param ) )
            print( "umin  ", fmt( umin ) )
            print( "umax  ", fmt( umax ) )
            print( "usav  ", fmt( usav ) )

        reset = True
        ptry = param.copy()
        step = 0

        for ks in range( nstep ) :

            ## orthonormalise the random vector
            vel = onb.normalise( vel, reset=reset )

            if self.verbose > 4 :
                print( ks, step, "vel   ", fmt( vel ) )

            ## append travel times wrt usav (==0) to all edges umin and umax
            ## tt contains entrance times as negative values
            ## and exit times as positive values.
            tt = numpy.append( ( umin - usav ) / vel, ( umax - usav ) / vel )

            ## find smallest entrance and exit times
            t1 = numpy.min( numpy.where( tt > 0, tt, +math.inf ) )
            t0 = numpy.max( numpy.where( tt < 0, tt, -math.inf ) )

            ## same now for the values at the edges of the unit box (0,1)
            if self.debug :
                tt = numpy.append( - usav / vel, ( 1.0 - usav ) / vel )
                t1max = numpy.min( numpy.where( tt > 0, tt, +math.inf ) )
                t0max = numpy.max( numpy.where( tt < 0, tt, -math.inf ) )

                t0 = self.stepOut( problem, ptry, usav, vel, t0, t0max, lowLhood, fitIndex )
                t1 = self.stepOut( problem, ptry, usav, vel, t1, t1max, lowLhood, fitIndex )

            kk = 0
            while True :
                kk += 1

                assert t0 < 0 < t1, "%f < %f < %f"%(t0,0,t1)
                ## dt is timestep wrt. usav (at 0)
                dt = t0 + self.rng.rand( 1 ) * ( t1 - t0 )
                utry = usav + vel * dt

                ptry[fitIndex] = self.unit2Domain( problem, utry, kpar=fitIndex  )

                Ltry = self.errdis.logLikelihood( problem, ptry )

                if self.verbose > 4 :
                    print( kk, fmt(t0), fmt(t1), fmt(dt), fmt(t1-t0), fmt(Ltry) )

                if Ltry >= lowLhood:
                    self.reportSuccess( )

                    step += 1

                    self.plotter.move( param, ptry, col=0, sym=0 )

                    ## check if better than Lbest in walkers[-1]
                    # self.checkBest( problem, ptry, Ltry, fitIndex )

                    ## keep the trial parameters
                    param = ptry.copy()
                    usav = self.domain2Unit( problem, param[fitIndex], kpar=fitIndex )

                    # make sure that 0 <= umin < usav < umax <= 1
                    umin = numpy.where( usav < umin, 2 * usav - umin, umin )
                    umax = numpy.where( usav > umax, 2 * usav - umax, umax )
                    umin = numpy.maximum( umin, 0 )
                    umax = numpy.minimum( umax, 1 )

                    ## find a new random direction, orthonormal to the previous ones
                    vel = self.rng.rand( np ) - 0.5

                    ## TBC; for now reset always.
                    reset = False

                    ## update the walker
                    update = len( self.walkers ) if append else kw
                    self.setWalker( update, problem, ptry, Ltry, fitIndex=fitIndex )

                    break

                elif kk > self.maxtrials :
                    self.reportFailed()

                    ## find new random direction; discard the previous ie. reset the basis
                    vel = self.rng.rand( np ) - 0.5
                    reset = True

                    break

                self.plotter.move( param, ptry, col=5, sym=4 )
                self.reportReject( )

                if dt < 0 :
                    t0 = dt
                else :
                    t1 = dt

        if step == 0 and self.verbose > 4 :
            warnings.warn( "ChordEngine: no steps found" )

        self.plotter.stop( param=param, name="ChordEngine" )

        return step


    def stepOut( self, problem, ptry, usav, vel, t, tmax, lowLhood, fitIndex ) :
        """
        Check if endpoints are indeed outside the lowLhood domain.
        """

        param = self.unit2Domain( problem, usav, kpar=fitIndex )

        ## Step out
        if t == 0 :
            t = 0.01 * tmax
        while True :
#            self.reportFailed()

            utry = usav + vel * t
            ptry[fitIndex] = self.unit2Domain( problem, utry, kpar=fitIndex  )
            Ltry = self.errdis.logLikelihood( problem, ptry )
            if Ltry <= lowLhood :
                self.plotter.move( param, ptry, col=4 )
                return t
            else :
                t *= 2
                if abs( t ) >= abs( tmax ) :
                    utry = usav + vel * tmax
                    ptry[fitIndex] = self.unit2Domain( problem, utry, kpar=fitIndex  )
                    self.plotter.move( param, ptry, col=4 )
                    return tmax

        return t







import numpy as numpy
from astropy import units
import math

from . import Tools
from .Tools import setAttribute as setatt
from .Formatter import formatter as fmt

from .LinearModel import LinearModel
from .Dynamic import Dynamic
from .Modifiable import Modifiable
from .ExponentialPrior import ExponentialPrior
from .UniformPrior import UniformPrior

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"


#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *    2019 - 2020 Do Kester


class DecisionTreeModel( Modifiable, Dynamic, LinearModel ):
    """
    A DecisionTree Model (DTM) is mostly defined on multiple input dimensions (axes).
    It splits the data in 2 parts, according low and high values on a certain input axis.
    The splitting can continue along other axes.

    The axes can contain float values, categorials (int) or booleans.
    The float axes can have Nans when data are unkown.
    Each category set one bit in an integer. The unknown category is a category
    by itself.
    Booleans that contain unknowns should be coded as categorial.

        f( x:p ) = DTM
                    \left => DTM (or None)
                    \rite => DTM (or None)
                    \dimension
                    \split or \mask

    The tree is searched left to right.

    The parameters are all initialized at 0.0

    Examples
    --------
    >>> dtm = DecisionTreeModel( )
    >>> print( dtm )
    DecisionTree: with 0 components and 1 parameters

    Attributes
    ----------
    left : None or DTM
        a None the tree stops otherwise there is a new split.
    rite : None or DTM
        a None the tree stops otherwise there is a new split.
    dimension : int
        split according to data values on this axes
    itype : char. Either 'f' (float), 'i' (integer), or 'b' (boolean)
        characterizes the input dimension as float, integer or boolean
    split : float between 0 and 1 (for float dimension)
        to the left normalized values < split; to the rite normvalues > split.
        Unknown values (NaNs to to the smallest faction).
    nsplit : int
        number of calls to random.rand() to be averaged. Prior on "split"
    mask : float between 0 and 1 (for float dimension)
        to the left normalized values < split; to the rite normvalues > split.
        Unknown values (NaNs to to the smallest faction.

    Attributes from Modifiable
    --------------------------
        modifiable

    Attributes from Dynamic
    -----------------------
        dynamic

    Attributes from Model
    ---------------------
        npchain, parameters, stdevs, xUnit, yUnit

    Attributes from FixedModel
    --------------------------
        npmax, fixed, parlist, mlist

    Attributes from BaseModel
    --------------------------
        npbase, ndim, priors, posIndex, nonZero, tiny, deltaP, parNames

    """

    NSPLITPRIOR = 3


    def __init__( self, ndim=1, depth=0, split=0.5, kdim=0, itypes=[0], modifiable=True,
                        dynamic=True, code=None, growPrior=None, copy=None, **kwargs ):
        """
        DecisionTree model.

        The DTM standardly has a UniformPrior for all parameters, with limits [0,1]

        Parameters
        ----------
        ndim : int
            number of input dimensions
        depth : int
            depth of the tree
        kdim : None or list of ints
            input channel to be splitted.
        itypes : [list of] int
            indicating the type of input: 0 for float, 1 for boolean, >1 for categorial
            The last number is repeated for remaining inputs
        split : None or float or list of floats
            fraction (0<s<1) of the input kdim that falls on either side.
        growPrior : None or Prior
            Governing the growth
        modifiable : bool (True)
            Will the model modify dimension and/or split/mask
        dynamic : bool (True)
            Will the model grow/shrink

        """

        Modifiable.__init__( self, modifiable=modifiable )
        Dynamic.__init__( self, dynamic=dynamic )

        if depth == 0 or code is not None :
            nparams = 1
            setatt( self, "left", None )
            setatt( self, "rite", None )
            setatt( self, "split", None )
            setatt( self, "parent", None )

        else :
            d = depth - 1
            setatt( self, "dimension", kdim[0] )
            itype = Tools.getItem( itypes, self.dimension )

            klen = Tools.length( kdim )
            if not isinstance( split, list ) : split = [split]
            for k in range(  Tools.length( split ), klen ) :
                split += [Tools.getItem( split, k )]

            self.setSplitOrMask( itype, split[0] )

            kd1 = kdim[1:]
            sp1 = split[1:]
            left = DecisionTreeModel( ndim=ndim, depth=d, kdim=kd1, split=sp1,
                                      itypes=itypes, **kwargs )
            setatt( self, "left", left )
            setatt( self.left, "parent", self )

            ns = ( 0x1 << d ) - 1
            kd2 = kd1[ns:]
            sp2 = sp1[ns:]
            rite = DecisionTreeModel( ndim=ndim, depth=d, kdim=kd2, split=sp2,
                                      itypes=itypes, **kwargs )
            setatt( self, "rite", rite )
            setatt( self.rite, "parent", self )
            nparams = self.left.npars + self.rite.npars


        param = [0.0] * nparams

        LinearModel.__init__( self, nparams, ndim=ndim, copy=copy,
                    params=param, **kwargs )

        if copy is None :
            if growPrior is None :
                growPrior = ExponentialPrior( scale=2 )
            setatt( self, "growPrior", growPrior )
            setatt( self, "itypes", itypes )
            setatt( self, "ncomp", ( 0x1 << depth ) - 1 )
#            setatt( self, "nsplit", 3 )
#            self.setLimits( 0, 1 )
            setatt( self, "nsplit", self.NSPLITPRIOR )
            self.setLimits( lowLimits=0.0, highLimits=1.0 )
        else :
            setatt( self, "growPrior", copy.growPrior )
            setatt( self, "ncomp", copy.ncomp )
            setatt( self, "nsplit", copy.nsplit )
            setatt( self, "itypes", copy.itypes )

        gen = self.walk()
        child = next( gen )     # skip root
        while True :
            try :
                child = next( gen )
            except :
                break
            if child.itypes is not None :
                del( child.itypes )
                del( child.priors )

        if code is not None :
            self = self.decode( code, kdim, split, 0 )


    def copy( self ):
        """ Copy method.  """

        dtm = DecisionTreeModel( copy=self, modifiable=self.modifiable )

        if self.dimension is not None :
            setatt( dtm, "dimension", self.dimension )
            if self.split is not None :
                setatt( dtm, "split", self.split )
            else :
                setatt( dtm, "mask", self.mask )

            if hasattr( self, "itypes" ) :
                setatt( dtm, "itypes", self.itypes )

            setatt( dtm, "left", self.left.copy() )
            setatt( dtm.left, "parent", dtm )
            setatt( dtm, "rite", self.rite.copy() )
            setatt( dtm.rite, "parent", dtm )

        return dtm

    def __getattr__( self, name ) :
        """
        Return None when the named attribute has not been set.

        Parameters
        ----------
        name : string
            name of the attribute
        """
        if ( ( name == "split" ) or ( name == "mask" ) or ( name == "itypes" ) or
             ( name == "dimension" ) or ( name == "maxComp" ) or ( name == "minComp" ) or
             ( name == "parent" ) ) :
            return None
        else :
            return super().__getattr__( name )

    def setSplitOrMask( self, itype, split ) :
        """ For internal use only """

        if itype == 0 :         ## float
            setatt( self, "split", split )
        else :                  ## boolean (1) or categorial (>1)
            setatt( self, "mask", int( split * ( 0x1 << itype ) + 1 ) )

    def isLeaf( self ) :
        """
        Return true if self is a leaf
        """
        return self.dimension is None

    def partitionList( self, xdata, plist ) :
        """
        Partition the xdata in plist over 2 new lists according to the DTM-branch.

        Paramaters
        ----------
        xdata : array-like
            the xdata
        plist : array of ints
            indices in xdata

        Return
        ------
        pl1, pl2 : 2 lists
            together containing all indices in plist
        """
        if len( plist ) == 0 :
            return ( [], [] )

        xd = Tools.getColumnData( xdata, self.dimension )[plist]        ## working data

        if self.split is not None :
            mn = numpy.min( xd )
            mx = numpy.max( xd )

            split = mn + ( mx - mn ) * self.split
            ### add NAN values to the smallest fraction
            cmp = xd > split if self.split < 0.5 else numpy.logical_not( xd <= split )
        elif self.mask >= 2 :
            cmp = ( xd & self.mask ) != 0
        else :
            cmp = ( xd == self.mask )

        pl1 = plist[numpy.where( cmp )]
        pl2 = plist[numpy.where( numpy.logical_not( cmp ) )]

        return ( pl1, pl2 )

    def baseResult( self, xdata, params ):
        """
        Returns the result of the model function.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        ndata = Tools.length( xdata )
        plist = numpy.arange( ndata, dtype=int )

#        print( params )
        res = numpy.zeros( ndata, dtype=type( params[0] ) )

        kpar = 0
        if self.isLeaf() :
            return params[0]

        res, kpar = self.recursiveResult( xdata, params, kpar, plist, res )

        return res

    def recursiveResult( self, xdata, params, kpar, plist, res ) :
        """ For internal use only """

        pl1, pl2 = self.partitionList( xdata, plist )

        if self.left.isLeaf() :
            res[pl1] = params[kpar]
            kpar += 1
        else :
            res, kpar = self.left.recursiveResult( xdata, params, kpar, pl1, res )

        if self.rite.isLeaf() :
            res[pl2] = params[kpar]
            kpar += 1
        else :
            res, kpar = self.rite.recursiveResult( xdata, params, kpar, pl2, res )

        return res, kpar

    def basePartial( self, xdata, params, parlist=None ):
        """
        Returns the partials at the input value.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials
        params : array_like
            values for the parameters.
        parlist : array_like
            list of indices active parameters (or None for all)

        """
        ndata = Tools.length( xdata )
        plist = numpy.arange( ndata, dtype=int ) if parlist is None else parlist

        kpar = 0
        if self.isLeaf() :
            return numpy.ones( (ndata,1), dtype=float )

        part = numpy.zeros( (ndata, self.npbase), dtype=float )

        part, kpar = self.recursivePartial( xdata, kpar, plist, part )

        return part

    def recursivePartial( self, xdata, kpar, plist, part ) :
        """ For internal use only """

        pl1, pl2 = self.partitionList( xdata, plist )

        if self.left.isLeaf() :
            part[pl1,kpar] = 1.0
            kpar += 1
        else :
            part, kpar = self.left.recursivePartial( xdata, kpar, pl1, part )

        if self.rite.isLeaf() :
            part[pl2,kpar] = 1.0
            kpar += 1
        else :
            part, kpar = self.rite.recursivePartial( xdata, kpar, pl2, part )

        return part, kpar

    def sortXdata( self, xdata ):
        """
        Reorder the xdata according to the parameter ordering

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the partials

        """
        ndata = Tools.length( xdata )
        plist = numpy.arange( ndata, dtype=int )

        if self.isLeaf() :
            return plist

        return numpy.asarray( self.recursiveOrder( xdata, plist ), dtype=int )

    def recursiveOrder( self, xdata, plist ) :
        """ For internal use only """

        pl1, pl2 = self.partitionList( xdata, plist )

        if not self.left.isLeaf() :
            pl1 = self.left.recursiveOrder( xdata, pl1 )

        if not self.rite.isLeaf() :
            pl2 = self.rite.recursiveOrder( xdata, pl2 )

        return numpy.append( pl1, pl2 )


    def baseDerivative( self, xdata, params ) :
        """
        Return the derivative df/dx at each xdata (=x).

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        params : array_like
            values for the parameters.

        """
        return numpy.zeros( Tools.length( xdata ), dtype=float )

    def baseName( self ):
        """
        Returns a string representation of the model.

        """
        cl, cb = self.count()
        name = str( "DecisionTree: with %d components and %d parameters" % ( cb, cl ) )
        return name

    def fullName( self, ids=False ):
        """
        Returns a string representation of the model.

        Parameters
        ----------
        ids : bool
            if True give the pointers of the links too.

        """

        name = str( "DecisionTree:" )

        indent = 4
        kpar = 0
        name, kpar = self.recursiveName( name, indent, kpar, ids=ids )
        return name


    def recursiveName( self, name, indent, kpar, ids=False ) :
        """ For internal use only """
        if ids :
            if self.parent is None :
                name += " (%d)" % id( self )
            else :
                name += " (%d %d)" % ( id( self ), id( self.parent ) )

        if self.isLeaf() :
            name += " parameter %d\n" % kpar
            kpar += 1
        else :
            if self.split is not None :
                name += " dim %d split at %.2f\n" % ( self.dimension, self.split )
            else :
                name += " dim %d mask " % ( self.dimension )
                name += "None" if self.mask is None else numpy.binary_repr( self.mask )
                name += "\n"
            sind = " " * indent
            indent += 4
            name += "%sLeft  :" % sind
            name, kpar = self.left.recursiveName( name, indent, kpar, ids=ids )
            name += "%sRight :" % sind
            name, kpar = self.rite.recursiveName( name, indent, kpar, ids=ids )
        return ( name, kpar )


    def baseParameterUnit( self, k ):
        """
        Return the unit of the indicated parameter.

        Parameters
        ---------
        k : int
            parameter number.

        """
        return self.yUnit


    def walk( self ):
        """
        Iterate tree in pre-order depth-first search order

        Found this piece of code on the internet. Fairly obscure.
        """
        yield self
        for child in [self.left, self.rite]:
            if child is not None :
                for chld in child.walk( ):
                    yield chld


    def encode( self ) :
        """
        Make a code tuple to be used by the constructor to resurrect the DTM
        The tuple consists of
        code : list of (list or 1 or 2), encoding the structure of DTM
        dims : list of dimensions at the branches
        splim : list of split/mask values at the branches

        """
        if self.isLeaf() :
            return ( 1, [], [] )

        if self.left.isLeaf() and self.rite.isLeaf() :
            sm = self.split if self.split is not None else self.mask
            return ( 2, [self.dimension], [sm] )

        if self.left.isLeaf() :
            ( c, d, s ) = self.rite.encode()
            sm = self.split if self.split is not None else self.mask
            return ( [1,c], [self.dimension] + d, [sm] + s )

        if self.rite.isLeaf() :
            ( c, d, s ) = self.left.encode()
            sm = self.split if self.split is not None else self.mask
            return ( [c,1], [self.dimension] + d, [sm] + s )

        (c1,d1,s1) = self.left.encode()
        (c2,d2,s2) = self.rite.encode()
        sm = self.split if self.split is not None else self.mask

        return ( [c1,c2], [self.dimension] + d1 + d2, [sm] + s1 + s2 )

    def decode( self, code, kdim, splim, kbr ) :
        """
        Resurrect the DTM from the code generated by encode().
        For internal use in the Constructor.

        Parameters
        ----------
        code : list
            of (list or 1 or 2), encoding the structure of DTM
        dims : list
            of dimensions at the branches
        splim : list
            of split/mask values at the branches
        kbr : int
            counter (start at 0)
        """

        if code == 1 :
            return self, kbr-1

        dtm = self
        if code == 2 :
            setatt( dtm, "left", DecisionTreeModel( ndim=self.ndim, depth=0 ) )
            setatt( dtm, "rite", DecisionTreeModel( ndim=self.ndim, depth=0 ) )
            setatt( dtm, "dimension", kdim[kbr] )
            sname = "mask" if isinstance( splim[kbr], int ) else "split"
            setatt( dtm, sname, splim[kbr] )
            setatt( dtm.left, "parent", dtm )
            setatt( dtm.rite, "parent", dtm )

            return dtm, kbr

        setatt( dtm, "dimension", kdim[kbr] )
        sname = "mask" if isinstance( splim[kbr], int ) else "split"
        setatt( dtm, sname, splim[kbr] )

        left = DecisionTreeModel( ndim=self.ndim, depth=0 )
        left, kbr = left.decode( code[0], kdim, splim, kbr+1 )
        setatt( dtm, "left", left )
        setatt( dtm.left, "parent", dtm )

        rite = DecisionTreeModel( ndim=self.ndim, depth=0 )
        rite, kbr = rite.decode( code[1], kdim, splim, kbr+1 )
        setatt( dtm, "rite", rite )
        setatt( dtm.rite, "parent", dtm )

        return dtm, kbr


    def findLeaf( self, kleaf ) :
        """
        Find a leaf in the tree, returning the leaf.

        Parameter
        ---------
        kleaf : int
            index of the leaf to be found
        """
        kl = -1
        gen = self.walk()
        child = None
        while kl < kleaf :
            child = next( gen )
            if child.isLeaf() :
                kl += 1

        return child

    def findBranch( self, kbranch ) :
        """
        Find a branch in the tree, returning the branch.

        Parameter
        ---------
        kbranch : int
            index of the branch to be found
        """
#        print( self.fullName() )

#        print( "kbr   ", kbranch, self.ncomp, self.npars, self.count() )
        kl = -1
        gen = self.walk()
        child = None
        while kl < kbranch :
#            print( "kl   ", kl )
            child = next( gen )
            if not child.isLeaf() :
                kl += 1

        return child

    def check( self ) :
        """
        Find a branch in the tree, returning the branch.

        Parameter
        ---------
        kbranch : int
            index of the branch to be found
        """
        kbranch = self.countBranch()

        kl = -1
        gen = self.walk()
        child = None
        ok = True
        while kl < kbranch :
            try :
                child = next( gen )
            except :
                break
            if not child.isLeaf() :
                kl += 1
                sp = child.split
                mk = child.mask
                kd = child.dimension
                tp = Tools.getItem( self.itypes, kd )
                if sp is not None :
                    ok = ok and tp == 0 and ( 0 < sp < 1 )
                elif mk is not None :
                    ok = ok and tp > 0 and ( 0 < mk < ( 0x1 << tp + 1 ) )
                else :
                    ok = False

        return ok

    def findRoot( self ) :
        """
        Return the root of the tree.

        """
        root = self
        while root.parent is not None :
            root = root.parent

        return root

    def count( self ) :
        """
        Return number of leafs and branches.
        """
        cleaf = 0
        cbran = 0
        gen = self.walk()
        while True :
            try :
                child = next( gen )
            except :
                break

            if child.isLeaf() :
                cleaf += 1
            else :
                cbran += 1
        return cleaf, cbran

    def countLeaf( self ) :
        """
        Return number of leafs.
        """
        return self.count()[0]

    def countBranch( self ) :
        """
        Return number of leafs.
        """
        return self.count()[1]


    def grow( self, offset=0, rng=None, location=0, split=0.5, kdim=0 ):
        """
        Increase the the number of components by 1 (if allowed by maxComp)

        Parameters
        ----------
        offset : int
            offset in the parameter list (pertaining to earlier models in the chain)
        rng : Random Number Generator
            to obtain random values for items below.
        location : int
            location where the new dtm-leaf should be inserted
        kdim : int (<self.ndim)
            dimension to split
        split : float (0<split<1)
            relative cut on this dimension

        Return
        ------
        bool :  succes

        """
        if not self.isDynamic() :
            return False

        if self.maxComp is not None and self.ncomp >= self.maxComp:
            return False

        if rng is not None :
            location = rng.randint( self.npbase )
            kdim = rng.randint( self.ndim )
            itype = Tools.getItem( self.itypes, kdim )
            split = numpy.mean( rng.rand( self.nsplit ) ) if itype == 0 else rng.rand()
        else :
            itype = Tools.getItem( self.itypes, kdim )

        ## Replace the leaf by a DTM at its parent
        leaf = self.findLeaf( location )
        setatt( leaf, "dimension", kdim )

        leaf.setSplitOrMask( itype, split )

#        print( "GROW   ", kdim, itype, split, leaf.split, leaf.mask )

        setatt( leaf, "left", DecisionTreeModel( ndim=self.ndim, depth=0 ) )
        setatt( leaf.left, "parent", leaf )
        setatt( leaf, "rite", DecisionTreeModel( ndim=self.ndim, depth=0 ) )
        setatt( leaf.rite, "parent", leaf )

        ## count leafs (parameters) and branches (components)
        cleaf, cbran = self.count()
        dnp = cleaf - self.npars

        ## duplicate the parameter
        value = self._head.parameters[offset+location]
        self.alterParameterSize( dnp, offset, location, value=value )

        head = self._head
        mdlpar = head.parameters
        k1 = location + offset
        value = numpy.zeros( dnp, dtype=float )
        if rng is not None :
            for k in range( dnp ) :
                value[k] = head.unit2Domain( rng.rand(), k + k1 )

        mdlpar = self.alterParameters( mdlpar, location, dnp, offset, value=value )

        setatt( self._head, "parameters", mdlpar )

        self.alterParameterNames( dnp )

        setatt( self, "ncomp", cbran )

        return True

    def shrink( self, offset=0, rng=None, location=0 ):
        """
        Decrease the the number of componenets by 1 (if allowed by minComp)
        Remove an arbitrary item.

        Parameters
        ----------
        offset : int
            location where the new params should be inserted

        Return
        ------
        bool : succes

        """
        if self.ncomp < 1 or ( self.minComp is not None and self.ncomp <= self.minComp ):
            return False

        if rng is not None :
            location = rng.randint( self.npbase )

        leaf = self.findLeaf( location )
        parent = leaf.parent
#       npp = parent.countLeaf()

        if parent is None :
            return False


        ## Erase the branch by putting the leafs to None
        setatt( parent, "left", None )
        setatt( parent, "rite", None )
        setatt( parent, "split", None )
        setatt( parent, "dimension", None )

        cleaf, cbran = self.count()
        dnp = cleaf - self.npars
        location = min( location - dnp, self.npars )
        self.alterParameterSize( dnp, offset, location=location )

        head = self._head
        mdlpar = head.parameters

        mdlpar = self.alterParameters( mdlpar, location, dnp, offset )
        setatt( self._head, "parameters", mdlpar )

        self.alterParameterNames( dnp )

        setatt( self, "ncomp", cbran )

        return True

    def vary( self, rng=None, location=0, split=0.5, kdim=0 ):
        """
        Vary the model structure by changing kdim and/or split at location

        Parameters
        ----------
        rng : Random Number Generator
            to obtain random values for items below.
        location : int
            location where the dtm-branch should be changed
        kdim : int (<self.ndim)
            dimension to split
        split : float (0<split<1)
            relative cut on this dimension

        Return
        ------
        bool :  succes

        """
        if self.ncomp < 1 :
            return False

        if rng is not None :
            location = rng.randint( self.ncomp )
            parent = self.findBranch( location )
            if rng.rand() < 0.5 :
                kdim = rng.randint( self.ndim )
            else :
                kdim = parent.dimension

            itype = Tools.getItem( self.itypes, kdim )
            if itype == 0 :
                if parent.split is None :               ## make new split
                    split = numpy.mean( rng.rand( self.nsplit ) )
                else :                                  ## shift nearby
                    split = ( self.nsplit * parent.split + rng.rand() ) / ( self.nsplit + 1 )
                setatt( parent, "split", split )
            elif itype == 1 :                           ## make new mask
                if parent.mask is None :
                    mask = rng.randint( 2 ) + 1
                else :                                  ## exchange bits
                    mask = 2 if parent.mask == 1 else 1
                setatt( parent, "mask", mask )
            else :
                if parent.mask is None :                ## make new mask
                    mask = rng.randint( 0x1 << itype ) + 1
                else :                                  ## change 1 bit
                    kr = rng.randint( itype )
                    mask = parent.mask ^ ( 0x1 << kr )
                setatt( parent, "mask", mask )

        else :
            parent = self.findBranch( location )
            itype = Tools.getItem( self.itypes, kdim )
            parent.setSplitOrMask( itype, split )

        setatt( parent, "dimension", kdim )

        return True






# run with : python3 -m unittest TestSalesmanProblem

import unittest
import os
import numpy as numpy
from astropy import units
from numpy.testing import assert_array_equal as assertAE
import math
import matplotlib.pyplot as plt

from BayesicFitting import *
from BayesicFitting import formatter as fmt
from BayesicFitting import fma


__author__ = "Do Kester"
__year__ = 2018
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class Test( unittest.TestCase ):
    """
    Test harness for solving the SalesmanProblem with NestedSampler.

    Author       Do Kester

    """
    def __init__( self, name ):
        super( ).__init__( name )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )


    def makeData( self, np ) :
        n2 = np * np
        x0 = numpy.arange( n2, dtype=int ) % np
        x1 = numpy.arange( n2, dtype=int ) // np
        xdata = numpy.append( x0, x1 ).reshape( 2, n2 ).transpose()
        return numpy.array( xdata, dtype=float )
        
    def testSolve1( self, plot=False ):
        print( "\n   NestedSolver Test1 \n" )
        np = 6

        xdata = self.makeData( np )

        problem = SalesmanProblem( xdata )
        print( "mindis  ", problem.scale )

        print( problem )


        ns = NestedSolver( problem, verbose=2, seed=130105 )
#        ns.minimumIterations = 20000
#        ns.end = 4
#        ns.ensemble = 10

        sl = ns.solve()


        if self.doplot :
            pars = sl[-1].allpars

            x = problem.xdata[pars,0]
            y = problem.xdata[pars,1]
            print( x )
            print( y )

            plt.plot( x, y, 'k-' )
            plt.show()


    def testSolve2( self, plot=False ):
        print( "\n   NestedSolver Test2 \n" )
        np = 6

        n2 = np * np
        numpy.random.seed( 123456 )
        x1 = ( numpy.random.rand( n2 ) - 0.5 ) * 180
        x0 = numpy.random.rand( n2 ) * 360 - 180

#        x0 = numpy.arange( n2, dtype=int ) % np
#        x1 = numpy.arange( n2, dtype=int ) // np
#        x0 = ( x0 - hp ) * 180 / ( hp + 1 )
#        x1 = ( x1 - hp ) * 90 / ( hp + 1 )

        print( x0 )
        print( x1 )

        xdata = numpy.append( x0, x1 ).reshape( 2, n2 ).transpose()
        xdata = numpy.array( xdata, dtype=float )

        problem = SalesmanProblem( xdata, distance="spheric" )
        print( "mindis  ", problem.scale )

        engs = ["move", "reverse", "switch"]

        ns = NestedSolver( problem, engines=engs, verbose=2, seed=80409 )

        sample = ns.solve()

#        print( fmt( problem.xdata, max=None ) )
        pars = sample.allpars
#        print( fmt( problem.xdata[pars], max=None ) )


        if self.doplot :

            d2r = math.pi / 180
            y = problem.xdata[pars,1]
            t = problem.xdata[pars,0]
            x = t * numpy.cos ( y * d2r )

#            print( x )
#            print( y )

            yy = numpy.linspace( -90, 90, 181 )
            cy = numpy.cos( yy * d2r )
            for k in range( -180, 181, 30 ) :
                plt.plot( k * cy, yy, 'g-' )

            for k in range( -90, 90, 30 ) :
                xx = 180 * math.cos( k * d2r )
                plt.plot( [-xx, xx], [k, k], 'g-' )

            rt = numpy.roll( t, -1 )
            rx = numpy.roll( x, -1 )
            ry = numpy.roll( y, -1 )

            for k in range( n2 ) :
                if x[k] < rx[k] :
                    x0,x1,t0,t1,y0,y1 = (x[k],rx[k],t[k],rt[k],y[k],ry[k]) 
                else :
                    x0,x1,t0,t1,y0,y1 = (rx[k],x[k],rt[k],t[k],ry[k],y[k])

                tm = 180
                if ( t1 - t0 ) < tm :
                    plt.plot( [x0,x1], [y0,y1], 'k-' )
                else :
                    ym = y0 + ( y1 - y0 ) * ( tm - t1 ) / ( 360 - ( t1 - t0 ) )
                    xm = tm * math.cos( d2r * ym )
                    print( fmt( [x0, xm, x1] ), fmt( [y0,ym,y1] ) ) 
                    plt.plot( [x1,xm], [y1,ym], 'r-' )
                    plt.plot( [x0,-xm], [y0,ym], 'b-' )

#            plt.plot( x, y, 'k-' )
            plt.show()



    def suite( cls ):
        return unittest.TestCase.suite( TestNestedSolver.__class__ )


if __name__ == '__main__':
    unittest.main( )





# run with : python3 -m unittest TestAstropyModels.Test.test1

import unittest
import os
import numpy as numpy
from astropy import units
from astropy import modeling
#from astropy.modeling.models import Gaussian1D
#from astropy.modeling.models import Polynomial1D
import matplotlib.pyplot as plt
import warnings

from StdTests import stdModeltest

from BayesicFitting import *
from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class Test( unittest.TestCase ):
    """
    Test harness for Models

    Author:      Do Kester

    """
    def __init__( self, testname ):
        super( ).__init__( testname )
        self.doplot = ( "DOPLOT" in os.environ and os.environ["DOPLOT"] == "1" )

    def test1a( self ):
        print( "******ASTROPY MODEL 1a**********************" )
        gm = modeling.models.Gaussian1D()

#        m = AstropyModel( gm )
        p = numpy.asarray( [1.2, -0.1, 0.3], dtype=float )

#        stdModeltest( m, p, plot=self.doplot )

        x = numpy.linspace( -2, 2, 6, dtype=float )
#        print( gm )
        gr = gm.evaluate( x, *tuple( p ) )
        gp = gm.fit_deriv( x, *tuple( p ) )

        print( gr.shape )
        print( fmt( gr ) )
        print( gp.__class__, gp[0].shape )
        print( fmt( gp, max=None ) )


        hm = modeling.models.Hermite1D( 2 )
        hr = hm.evaluate( x, *tuple( p ) )
        hp = hm.fit_deriv( x, *tuple( p ) )

#        print( hm )
        print( hr.shape )
        print( fmt( hr ) )
        print( hp.__class__, hp[0].shape )
        print( fmt( hp, max=None ) )
                
    def test1b( self ):
        print( "******ASTROPY MODEL 1B**********************" )

        p = numpy.asarray( [1.2, -0.1, 0.3], dtype=float )
        x = numpy.linspace( -2, 2, 6, dtype=float )

        gm = modeling.models.Gaussian1D()
        m1 = AstropyModel( gm )

        gr = m1.result( x, p )
        gp = m1.partial( x, p )

        print( gr.shape )
        print( fmt( gr ) )
        print( gp.__class__, gp.shape )
        print( fmt( gp, max=None ) )


        hm = modeling.models.Hermite1D( 2 )
        m2 = AstropyModel( hm )
        hr = m2.result( x, p )
        hp = m2.partial( x, p )

#        print( hm )
        print( hr.shape )
        print( fmt( hr ) )
        print( hp.__class__, hp.shape )
        print( fmt( hp, max=None ) )
                


    def test2( self ):
        print( "******ASTROPY MODEL 2***********************" )
        gm = modeling.models.Gaussian1D()

        m = AstropyModel( gm, fixed={0:2.6} )
        p = numpy.asarray( [-0.1,30], dtype=float )

        stdModeltest( m, p, plot=self.doplot )

    def test3( self ):
        print( "******ASTROPY MODEL 3***********************" )
        gm = modeling.models.Gaussian1D()

        print( gm.__class__ )
        print( gm )

        m = AstropyModel( gm ) + PolynomialModel( 1 )
        p = numpy.asarray( [1.2,-0.1,30,1.0,0.3], dtype=float )

        stdModeltest( m, p, plot=self.doplot )

    def test4( self ):
        print( "******ASTROPY MODEL 4***********************" )
        gm = modeling.models.Gaussian1D()

#        print( gm.__class__ )
#        print( gm )

        apm = AstropyModel( gm )
        apm.setPrior( 0, ExponentialPrior( scale=10 ) )     
        apm.setPrior( 1, UniformPrior( limits=[-10,10] ) )     
        apm.setPrior( 2, JeffreysPrior( limits=[0.1,10] ) )     

        bfm = PolynomialModel( 1 )
        bfm.setPrior( 0, UniformPrior( limits=[-5,5] ) )

        m = apm + bfm
        p = numpy.asarray( [1.2,-0.1,30,1.0,0.3], dtype=float )

        N = 201
        x = numpy.arange( N, dtype=float ) / 25 - 2

        y = ( x - 0.70 ) / 0.4
        y *= -0.5 * y
        y = 8.0 * numpy.exp( y )
        y += 0.2 * x + 1.0
        numpy.random.seed( 13456 )
        y += 0.5 * numpy.random.randn( N )

        ns = NestedSampler( x, m, y )
        ns.distribution.setLimits( [0.01, 10] )
        ns.verbose = 2

        evi = ns.sample( plot=self.doplot )
        print( "NS pars ", fmt( ns.parameters ) )
        print( "NS stdv ", fmt( ns.stdevs ) )
        print( "NS scal ", fmt( ns.scale ) )


    def test5( self ):
        print( "******ASTROPY MODEL 1***********************" )
        gm = modeling.models.Gaussian2D()

#        print( gm.param_names )
#        print( gm.parameters )

        m = AstropyModel( gm )
        p = numpy.asarray( [1.2, -0.1, 0.2, 3.0, 2.0, 0.0], dtype=float )

        x = numpy.linspace( -5, 5, 11, dtype=float )
        x = numpy.stack( (x, x) ).T
#        print( x )

        stdModeltest( m, p, x=x, plot=self.doplot )

    ## Next one does not work as a AP-CompoundModel is not and AP-FittableModel
    def XXXtest4( self ):
        print( "******ASTROPY MODEL 4***********************" )
        gm = Gaussian1D() + Polynomial1D( 1 )
        
        print( gm.parameters )
        print( gm )

        m = AstropyModel( gm )
        p = numpy.asarray( [1.2,-0.1,30,1.0,0.3], dtype=float )

        stdModeltest( m, p, plot=self.doplot )


    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestAstropyModel.__class__ )

if __name__ == '__main__':
    unittest.main( )





# run with : python3 -m unittest TestSampleList

import unittest
import numpy as numpy
import sys
from numpy.testing import assert_array_almost_equal as assertAAE
from astropy import units
import math

from BayesicFitting import *
from BayesicFitting import formatter as fmt

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2002 Do Kester

class TestSampleList( unittest.TestCase ):
    """
    Test harness for Fitter class.

    Author       Do Kester

    """

    # Define x independent variable
    x = numpy.asarray( [ -1.0, -0.8, -0.6, -0.4, -0.2, 0.0, 0.2, 0.4, 0.6, 0.8, 1.0], dtype=float )

    # Define noise: random Gaussian noise sig=0.3
    noise = numpy.asarray( [ -0.000996, -0.046035,  0.013656,  0.418449,  0.0295155,  0.273705,
    -0.204794,  0.275843, -0.415945, -0.373516, -0.158084], dtype=float )
    wgt = numpy.asarray( [1,2,3,4,5,6,7,4,3,2,1], dtype=float )                 #  total 38
    par = numpy.asarray( [3,2,1,0.3], dtype=float )
    len = 11


    #  **************************************************************
    def testSampleList( self ):
        print( "=========  SampleListTest  =======================" )
        gm = GaussModel( )
        gm += PolynomialModel( 0 )

        problem = ClassicProblem( gm, xdata=self.x, ydata=self.noise )
#        errdis = GaussErrorDistribution( )

        lnZ = 1.234
        sl0 = SampleList( gm, self.len )
        k = 0
        for s in sl0 :
            self.assertTrue( s.id == k )
            self.assertTrue( s.parent == -1 )
            self.assertTrue( isinstance( s.model, GaussModel ) )
            self.assertTrue( s.logW == 0 )
            self.assertTrue( s.logL == 0 )
            self.assertTrue( s.parameters[0] == 1 )
            self.assertTrue( s.parameters[1] == 0 )
            self.assertTrue( s.parameters[2] == 1 )
            self.assertTrue( s.parameters[3] == 0 )

            self.assertTrue( s.fitIndex is None or len( s.fitIndex ) == 4 )
            k += 1

        ap = numpy.append( gm.parameters, [0.5] )
        fi = numpy.asarray( [0,1,2,3,-1] )

        sl = SampleList( gm, self.len, parameters=ap, fitIndex=fi )
        k = 0
        for s in sl:
            s.id = k + 1
            s.parent = ( k + 2 ) % self.len + 1
            sup = 0.3 + 0.01 * self.noise[k]
            s.hyper = sup
            pars = self.par + 0.01 * self.noise[k]
            s.parameters = pars
            s.logL = -1213.0 + self.x[k]
            s.logW = math.log( self.wgt[k] / 38.0 ) + lnZ
            print( s )
            print( "    allpars ", fmt( s.allpars ) )
            print( "    par sup ", fmt( s.parameters ), fmt( s.hypars ) )
            print( "    fitindx ", fmt( s.fitIndex ) )
            k += 1
        sl.logZ = lnZ
        sl.info = 30
        self.assertTrue( sl.evidence == sl.logZ / math.log( 10) )
        self.assertTrue( sl.info == 30 )

        sumw = 0
        for s in sl :
            sumw += s.weight

        print( "wgt evo ", sl.getWeightEvolution( ) )
        print( "sum wgt ", sumw )
        self.assertFalse( abs(sumw - 1.0 ) < 1e-8 )

        sl.normalize()
        sumw = 0
        for s in sl :
            sumw += s.weight
        print( sl.getWeightEvolution( ) )
        print( sumw )
        self.assertTrue( abs(sumw - 1.0 ) < 1e-8 )

        for s in sl0 :
            s.logW = -sys.float_info.max

        sl0.sample( 4, sl.sample( 4 ) )
        sl0[5] = sl[5]
        sl0.logZ = sl.logZ

        print( sl0.sample( 4 ) )
        print( sl0[4].parameters )
        print( sl0[4].hypars )

        print( sl.sample( 4 ) )
        print( sl[4].parameters )
        print( sl[4].hypars )

        print( sl0.sample( 5 ) )
        print( sl.sample( 5 ) )

        self.assertTrue( sl0[4].id == sl[4].id )
        self.assertTrue( sl0[4].parent == sl[4].parent )
        self.assertTrue( sl0[4].hypars == sl[4].hypars )
        self.assertTrue( sl0[4].logW == sl[4].logW )
        self.assertTrue( sl0.sample( 4 ).logL == sl[4].logL )

        print( "sl0 Id   ", sl0.getGeneration( ) )
        print( "sl0 par  ", sl0.getParentEvolution( ) )
        print( "sl  Id   ", sl.getGeneration( ) )
        print( "sl  par  ", sl.getParentEvolution( ) )

        sl0.add( sl[10] )
#        print( "sl0 Id   ", sl0.getGeneration( ) )
        self.assertTrue( len( sl0 ) == 12 )
        sl0.copy( 11, 0 )
#        print( "sl0 Id   ", sl0.getGeneration( ) )
        self.assertTrue( sl0[0].id == 0 )
        self.assertTrue( sl0[11].id == 11 )
        self.assertTrue( sl0[0].logW == sl0[11].logW )
        self.assertTrue( sl0.sample( 0 ).logL == sl0.sample( 11 ).logL )

        for s,t in zip( sl0, sl ) :
            s.logL = t.logL
            s.logW = t.logW

        print( "SL   ", len( sl0 ) )
#        Tools.printclass( sl0[0] )
#        Tools.printclass( sl0[7] )
        for s in sl0 :
            print( s.id, s.logW, s.logL )

        k = 1
        while k < 3:
            sl0.weed( maxsize=8 )
            print( "SL0  ", k, len( sl0 ) )
            for s in sl0 :
                print( s.id, s.logW, s.logL )
#            Tools.printclass( sl0[0] )
#            Tools.printclass( sl0[7] )
            self.assertTrue( len( sl0 ) == 8 )
#            self.assertTrue( sl0[0].logL == sl0[5].logL )
            k += 1

        print( "SL   ", k, len( sl ) )
        for s in sl :
            print( s.id, s.logW, s.logL )

        print( "par    ", sl.parameters )
        print( "stdev  ", sl.stdevs )
        print( "hypars ", sl.hypars )
        print( "stdscl ", sl.stdevHypars )

        print( sl.medianIndex, sl.modusIndex, sl.maxLikelihoodIndex )

        sss = numpy.zeros( gm.npchain, dtype=float ) + 0.003

        assertAAE( sl.parameters, self.par, 2 )
        assertAAE( sl.standardDeviations, sss, 1 )
        assertAAE( sl.scale, 0.3, 2 )
        assertAAE( sl.stdevScale, 0.003, 1 )
        self.assertTrue( sl.medianIndex == 5 )
        self.assertTrue( sl.modusIndex == 6 )
        self.assertTrue( sl.maxLikelihoodIndex == -1 )
        assertAAE( sl.maxLikelihoodParameters, sl[10].parameters )
        self.assertTrue( sl.maxLikelihoodScale == sl[10].hypars[0] )
        assertAAE( sl.medianParameters, sl[5].parameters )
        self.assertTrue( sl.medianScale == sl[5].hypars[0] )
        assertAAE( sl.modusParameters, sl[6].parameters )
        self.assertTrue( sl.modusScale == sl[6].hypars[0] )

        param = sl.getParameterEvolution( )
        print( param.shape )
        self.assertTrue( param.shape[0] == 11 )
        self.assertTrue( param.shape[1] == 4 )
        par0 = sl.getParameterEvolution( 0 )
        par1 = sl.getParameterEvolution( 1 )
        par2 = sl.getParameterEvolution( 2 )
        par3 = sl.getParameterEvolution( 3 )
        nrp = sl.getNumberOfParametersEvolution( )

        for i in range( self.len ):
            assertAAE( param[i], self.par + 0.01 * self.noise[i] )
            assertAAE( param[i, 0], par0[i] )
            assertAAE( param[i, 1], par1[i] )
            assertAAE( param[i, 2], par2[i] )
            assertAAE( param[i, 3], par3[i] )
            self.assertTrue( nrp[i] == 4 )

        xx = numpy.arange( 10, dtype=float ) * 0.2
        yf1 = sl.average( xx )
        yf2 = gm.result( xx, sl.getParameters( ) )
        assertAAE( yf1, yf2, 5 )
        err = sl.monteCarloError( xx )
        assertAAE( err, numpy.zeros( 10, dtype=float ), 2 )
        zz = numpy.arange( 20, dtype=float ) * 0.2
        assertAAE( sl.monteCarloError( zz ), numpy.zeros( 20, dtype=float ), 2 )

    @classmethod
    def suite( cls ):
        return unittest.TestCase.suite( TestSampleList.__class__ )



if __name__ == '__main__':
    unittest.main()





import numpy as numpy
from threading import Thread

from .Engine import Engine
from .Formatter import formatter as fmt
from . import Tools

__author__ = "Do Kester"
__year__ = 2021
__license__ = "GPL3"
__version__ = "2.7.0"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2003 - 2014 Do Kester, SRON (Java code)
#  *    2017 - 2021 Do Kester


threadErrors = []

class Explorer( object ):
    """
    Explorer is a helper class of NestedSampler, which contains and runs the
    diffusion engines.

    It uses Threads to parallelise the diffusion engines.

    Attributes
    ----------
    walkers : WalkerList
        walkers to be explored
    engines : [engine]
        list of engines to be used
    errdis : ErrorDistribution
        to be used
    rng : numpy.random.RandomState
        random number generator
    rate : float (1.0)
        governs processing speed (vs precision)
    maxtrials : int (5)
        number of trials
    verbose : int (0)
        level of blabbering
    lowLhood : float
        present low likelihood level
    iteration : int
        counting explorer calls

    Author       Do Kester.

    """
    TWOP32 = 2 ** 32

    def __init__( self, ns, threads=False ):
        """
        Construct Explorer from a NestedSampler object.

        Parameters
        ----------
        ns : NestedSampler
            the calling NestedSampler. It provides the attributes.

        """
        self.walkers = ns.walkers
        self.engines = ns.engines
        self.errdis = ns.distribution
        self.rng = ns.rng
        self.rate = ns.rate
        self.maxtrials = ns.maxtrials
        self.verbose = ns.verbose
        self.threads = threads
        self.usePhantoms = hasattr( ns, "usePhantoms" ) and ns.usePhantoms

        self.selectEngines = self.allEngines    ## default: always use all engines
        for eng in self.engines :
            if hasattr( eng, "slow" ) :
                self.selectEngines = self.selEngines    ## use selection
                break

    def explore( self, worst, lowLhood, iteration ):
        """
        Explore the likelihood function, using threads.

        Parameters
        ----------
        worst : [int]
            list of walkers to be explored/updated
        lowLhood : float
            level of the low likelihood

        """
        self.iteration = iteration
        engines = self.selectEngines( iteration )

        if not self.threads :
            for kw in worst :
                self.exploreWalker( kw, lowLhood, engines, self.rng )

            return

        ## We have Threads
        explorerThreads = []
        self.lowLhood = lowLhood

        nrep = Engine.NCALLS
        for kw in worst :
            seed = self.rng.randint( self.TWOP32 )
            exThread = ExplorerThread( "explorer_%d"%kw, kw, self, engines, seed )
            exThread.start( )
            explorerThreads += [exThread]

        for thread in explorerThreads :
            thread.join( )

            for k,engine in enumerate( engines ) :
                nc = 0
                for i in range( nrep ) :
                    nc += thread.engines[k].report[i]
                    engine.report[i] += thread.engines[k].report[i]
                engine.report[nrep] += nc

        if len( threadErrors ) > 0: #check if there are any errors
            for e in threadErrors:
                print( e )
            raise Exception( "Thread Error" )

    def exploreWalker( self, kw, lowLhood, engines, rng ):
        """
        Move the walker around until it is randomly distributed over the prior and
        higher in logL then lowLhood

        Parameters
        ----------
        kw : int
            index in walkerlist, of the walker to be explored
        lowLhood : float
            minimum value for the log likelihood
        engine : list of Engine
            to be used
        rng : RandomState
            random number generator
        """
        if self.verbose >= 4 :
            print( "-------- Start exploreWalker ---------------------------------" )
#            self.checkWalkers()

        walker = self.walkers[kw]
        oldlogL = walker.logL

        self.errdis.lowLhood = lowLhood

#        maxmoves = len( walker.fitIndex ) / self.rate
        maxmoves = len( walker.allpars ) / self.rate
        maxtrials = self.maxtrials / self.rate

        moves = 0
        trials = 0

        while moves < maxmoves and trials < maxtrials :

            for engine in rng.permutation( engines ) :
                mv = engine.execute( kw, lowLhood, append=self.usePhantoms,iteration=self.iteration )
                moves += mv

                update = len( self.walkers ) - 1 if self.usePhantoms else kw
#                if self.verbose >= 4 or self.walkers[update].logL < lowLhood :

                if self.verbose >= 4 and mv > 0 :
                    wlkr = self.walkers[update]
                    print( "%4d %-15.15s %4d %10.3f %10.3f ==> %3d  %10.3f"%
                            ( trials, engine, update, lowLhood, oldlogL, moves,
                                wlkr.logL ) )
                    print( "IN       ", fmt( walker.allpars, max=None, linelength=200 ), len( walker.allpars ) )
                    print( "OUT      ", fmt( wlkr.allpars, max=None, linelength=200 ), len( wlkr.allpars ) )

                    if len( wlkr.allpars ) < len( wlkr.fitIndex ) :
                        raise ValueError( "Walker parameter %d fitIndex %d" %
                            ( len( wlkr.allpars ), len( wlkr.fitIndex ) ) )
                    oldlogL = wlkr.logL

                    ## check all walkers for consistency
                    self.logLcheck( walker )

            trials += 1

        if moves == 0 :
            self.logLcheck( self.walkers[kw] )

        if self.walkers[update].logL < lowLhood :
            raise Exception( "%10.3f < %10.3f" % ( self.walkers[kw].logL, lowLhood )  )

        return

    def selEngines( self, iteration ) :
        """
        Select engines with slowly changing parameters once per so many iterations.

        Parameter
        ---------
        iteration : int
            iteration number
        """
        engines = []
        for eng in self.engines :
            if not hasattr( eng, "slow" ) or ( iteration % eng.slow ) == 0 :
                engines += [eng]

        return engines

    def allEngines( self, iteration ) :
        """
        Always use all engines.

        Parameters
        ----------
        iteration : int
            iteration number
        """
        return self.engines

    def checkWalkers( self ) :
        for w in self.walkers :
            self.logLcheck( w )


    def logLcheck( self, walker ) :
        """
        Sanity check when no moves are found, if the LogL is still the same as the stored logL.

        Parameters
        ----------
        walker : Walker
            the one with the stored logL

        Raises
        ------
        ValueError at inconsistency.

        """
        walker.check( nhyp=self.errdis.nphypar )

        if walker.problem.model.npars < len( walker.allpars ) - self.errdis.nphypar :
            Tools.printclass( walker )
            print( "Iteration %4d %4d %10.3f  %10.3f" % ( self.iteration, walker.id, walker.logL, wlogL ) )
            print( fmt( walker.allpars, max=None ) )
            raise ValueError( "Inconsistency in length of modelparams (%d) and allpars (%d - %d)" %
                                ( walker.problem.model.npars, len( walker.allpars ), self.errdis.nphypar ) )


        wlogL = self.errdis.logLikelihood( walker.problem, walker.allpars )
        if wlogL != walker.logL :
            Tools.printclass( walker )
            print( "Iteration %4d %4d %10.3f  %10.3f" % ( self.iteration, walker.id, walker.logL, wlogL ) )
            print( fmt( walker.allpars, max=None ) )
            raise ValueError( "Inconsistency between stored logL %f and calculated logL %f" %
                                ( walker.logL, wlogL ) )

        for ki in walker.fitIndex :
#            if walker.fitIndex[ki] >= 0 :
            if ki < 0 :
                self.errdis.hyperpar[ki].prior.checkLimit( walker.allpars[ki] )
            elif ki < walker.problem.model.npars :
                walker.problem.model.getPrior( ki ).checkLimit( walker.allpars[ki] )


class ExplorerThread( Thread ):
    """
    One thread for the Explorer. It updates one walker.

    Attributes
    ----------
    id : int
        identity for thread
    walkers : WalkerList
        list of walkers
    rng : numpy.random.RandomState
        random number generator
    engines : [Engine]
        copy of the list of Engines of Explorer
    """

    global threadErrors

    def __init__( self, name, id, explorer, engines, seed ):
        super( ExplorerThread, self ).__init__( name=name )
        self.id = id
        self.explorer = explorer
        self.engines = [eng.copy() for eng in engines]
        self.rng = numpy.random.RandomState( seed )


    def run( self ):
        try :
            self.explorer.exploreWalker( self.id, self.explorer.lowLhood,
                                         self.engines, self.rng )
        except Exception as e :
            threadErrors.append( [repr(e) + " occurred in walker %d" % self.id] )
            raise






# run with : python3 -m unittest TestNNUtilities

import unittest
import numpy as numpy
from astropy import units
import math
import matplotlib.pyplot as plt
import matplotlib.cm as cm

from BayesicFitting import formatter as fmt

from BayesicFitting import *

__author__ = "Do Kester"
__year__ = 2017
__license__ = "GPL3"
__version__ = "0.9"
__maintainer__ = "Do"
__status__ = "Development"

#  *
#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  *  2006 Do Kester

class Test( unittest.TestCase ):
    """
    Test harness for NeuralNetUtilities.

    Author       Do Kester

    """
    def __init__( self, name ):
        super( ).__init__( name )


    def test1( self ):
        print( "======= Neural Net Utilities  test 1 ========================" )

        con = NeuralNetUtilities.Connect( ndim=2, ndout=3 )

        print( con.ndim, con.ndout, con.nraster  )

        self.assertTrue( con.nraster == 6 )

        xx = numpy.array( [[1,2], [2,3], [3,4], [4,5]], dtype=float )
        par = numpy.arange( 6, dtype=float ) + 1

        print( fmt( xx ) )
        print( fmt( par, max=None ) )

        res = con.result( xx, par )
        print( "result  ", res.shape )
        print( fmt( res ) )

        partial = con.partial( xx, par )
        print( "partial ", partial.shape )
        print( fmt( partial, max=None ) )
        print( fmt( partial[0], max=None ) )
        print( fmt( partial[1], max=None ) )
        print( fmt( partial[2], max=None ) )


        deriv = con.derivative( xx, par )
        print( "deriv   ", deriv.shape )
        print( fmt( deriv, max=None ) )


    def test2( self ):
        print( "======= Neural Net Utilities  test 2 ========================" )

        con = NeuralNetUtilities.ConnectWithBias( ndim=2, ndout=3 )

        print( con.ndim, con.ndout, con.nraster  )

        self.assertTrue( con.nraster == 6 )

        xx = numpy.array( [[1,2], [2,3], [3,4], [4,5]], dtype=float )
        par = numpy.arange( 9, dtype=float ) + 1

        print( fmt( xx ) )
        print( fmt( par, max=None ) )

        res = con.result( xx, par )
        print( "result  ", res.shape )
        print( fmt( res ) )

        partial = con.partial( xx, par )
        print( "partial ", partial.shape )
        print( fmt( partial, max=None ) )

        deriv = con.derivative( xx, par )
        print( "deriv   ", deriv.shape )
        print( fmt( deriv, max=None ) )

    def test3( self ):
        print( "======= Neural Net Utilities  test 3 ========================" )

        con = NeuralNetUtilities.ConnectWithBias( ndim=2, ndout=3 )
        atn = NeuralNetUtilities.Arctan()

        print( con.ndim, con.ndout, con.nraster  )

        self.assertTrue( con.nraster == 6 )

        xx = numpy.array( [[1,2], [2,3], [3,4], [4,5]], dtype=float )
        par = numpy.arange( 9, dtype=float ) + 1

        print( fmt( xx ) )
        print( fmt( par, max=None ) )

        res = atn.result( con.result( xx, par ), par )
        print( "result  ", res.shape )
        print( fmt( res ) )

        dfdr = atn.derivative( con.result( xx, par ), par )
        drdp = con.partial( xx, par )
        partial = atn.pipe( dfdr, drdp )
        print( "partial ", partial.shape )
        print( fmt( partial, max=None ) )

    def test4( self ):
        print( "======= Neural Net Utilities  test 4 ========================" )

        con = NeuralNetUtilities.ConnectWithBias( ndim=2, ndout=3 )
        atn = NeuralNetUtilities.Logistic()

        print( con.ndim, con.ndout, con.nraster  )

        self.assertTrue( con.nraster == 6 )

        xx = numpy.array( [[1,2], [2,3], [3,4], [4,5]], dtype=float )
        par = numpy.arange( 9, dtype=float ) + 1

        print( fmt( xx ) )
        print( fmt( par, max=None ) )

        res = atn.result( con.result( xx, par ), par )
        print( "result  ", res.shape )
        print( fmt( res ) )

        partial = atn.derivative( con.partial( xx, par ), par )
        print( "partial ", partial.shape )
        print( fmt( partial, max=None ) )

    def test5( self ):
        print( "======= Neural Net Utilities  test 5 ========================" )

        con = NeuralNetUtilities.ConnectWithBias( ndim=2, ndout=3 )
        atn = NeuralNetUtilities.Heaviside()

        print( con.ndim, con.ndout, con.nraster  )

        self.assertTrue( con.nraster == 6 )

        xx = numpy.array( [[1,2], [2,3], [3,4], [4,5]], dtype=float )
        par = numpy.arange( 9, dtype=float ) - 4

        print( fmt( xx ) )
        print( fmt( par, max=None ) )

        res = atn.result( con.result( xx, par ), par )
        print( "result  ", res.shape )
        print( fmt( res ) )

        dfdr = atn.derivative( con.result( xx, par ), par )
        drdp = con.partial( xx, par )
        partial = atn.pipe( dfdr, drdp )

        print( "partial ", partial.shape )
        print( fmt( partial, max=None ) )


    def test6( self ):
        print( "======= Neural Net Utilities  test 6 ========================" )

        con = NeuralNetUtilities.ConnectWithBias( ndim=2, ndout=3 )
        atn = NeuralNetUtilities.Softmax( ndout=3 )

        print( con.ndim, con.ndout, con.nraster  )

        self.assertTrue( con.nraster == 6 )

        xx = numpy.array( [[2,1], [2,3], [3,4], [4,5]], dtype=float )
        par = numpy.arange( 9, dtype=float )
        par[6:] = [0,-1,1]
        par[0],par[5] = par[5],par[0]

        print( fmt( xx ) )
        print( fmt( par, max=None ) )

        sm = SoftMaxModel( ndim=2, ndout=3, offset=True )

        res = con.result( xx, par )
        print( "result  con  ", res.shape )
        print( fmt( res ) )

        res = atn.result( res, par )
        print( "result  atn  ", res.shape )
        print( fmt( res ) )

        print( sm )
        smr = sm.result( xx, par )
        print( "result  sm   ", smr.shape )
        print( fmt( smr ) )


        dzdy = atn.derivative( con.result( xx, par ), par )
        print( "dfdr    ", dzdy.shape )
        print( fmt( dzdy, max=None ) )

        dydx = con.partial( xx, par )
        print( "drdp    ", dydx.shape )
        print( fmt( dydx, max=None ) )

        dzdx = numpy.zeros_like( dydx )

        for k in range( 9 ) :
            for i in range( 4 ) :
                dzdx[:,i,k] = numpy.inner( dzdy[:,i,:], dydx[:,i,k] )
        
#        dzdx = dzdy * dydx
        print( "dfdp    ", dzdx.shape )
        print( fmt( dzdx, max=None ) )

        printclass( sm )
        print( sm.trans.pipe )

        print( "SoftMax partial" )
        print( fmt( sm.partial( xx, par ), max=None ) )

        print( "SoftMax deriv  " )
        print( fmt( sm.derivative( xx, par ), max=None ) )










    def suite( cls ):
        return unittest.TestCase.suite( TestSplines.__class__ )


if __name__ == '__main__':
    unittest.main( )





import numpy as numpy
from astropy import units
import re
import warnings
from . import Tools
from .Tools import setAttribute as setatt

from .Prior import Prior
from .UniformPrior import UniformPrior

__author__ = "Do Kester"
__year__ = 2020
__license__ = "GPL3"
__version__ = "2.5.3"
__url__ = "https://www.bayesicfitting.nl"
__status__ = "Perpetual Beta"

#  * This file is part of the BayesicFitting package.
#  *
#  * BayesicFitting is free software: you can redistribute it and/or modify
#  * it under the terms of the GNU Lesser General Public License as
#  * published by the Free Software Foundation, either version 3 of
#  * the License, or ( at your option ) any later version.
#  *
#  * BayesicFitting is distributed in the hope that it will be useful,
#  * but WITHOUT ANY WARRANTY; without even the implied warranty of
#  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  * GNU Lesser General Public License for more details.
#  *
#  * The GPL3 license can be found at <http://www.gnu.org/licenses/>.
#  *
#  * A JAVA version of this code was part of the Herschel Common
#  * Science System (HCSS), also under GPL3.
#  *
#  *    2011 - 2014 Do Kester, SRON (Java code)
#  *    2016 - 2020 Do Kester

class BaseModel( object ):
    """
    BaseModel implements the common parts of simple Models.

    A simple model is a function in x with parameters p : f(x:p).
    The variable x is an array of points in a space of one or more
    dimensions; p can have any length, including 0 i.e. the model
    has no parameters.

    The result of the function for certain x and p is given by
    model.result( x, p )

    The partial derivatives of f to p (df/dp) is given by
    model.partial( x, p )
    Some fitters make use of the partials

    The derivative of f to x (df/dx) is given by
    model.derivative( x, p )

    BaseModel checks parameters for positivity and nonzero-ness, if such
    is indicated in the model itself.

    BaseModel also implements the numerical calculation of the (partial)
    derivatives to be used when they are not given in the model definition
    itself.

    Attributes
    ----------
    npbase : int
        number of params in the base model
    ndim : int
        number of dimensions (parallel streams) of input. (default : 1)
    priors : list of Prior
        pertaining to each of the parameters of the model.
        If the list is shorter than the number of parameters, the last one is repeated.
    posIndex : list of int
        list of indices indication positive-definite parameters.
    nonZero : list of int
        list of parameters that need a warning when they are equal to zero.
        Warnings will only be issued once. Values are replaced by self.tiny
    tiny : float
        very small value, replacing zero valued when found on NonZero.
        (default : 1e-20)
    deltaP : array_like
        (list of) width(s) for numerical partial calculation. (default : 0.00001)
    parNames : list of str
        list of parameter names. (default : "parameter_k")

    Author :         Do Kester

    """

    #  *************************************************************************
    def __init__( self, nparams=0, ndim=1, copy=None, posIndex=[], nonZero=[], **kwargs ):
        """
        BaseModel Constructor.
        <br>
        Parameters
        ----------
        nparams : int
            Number of parameters in the model (default: 0)
        ndim : int
            Number of dimensions of the input (default: 1)
        copy : BaseModel
            to be copied
        posIndex : list of int
            indices of parameters that need to be > 0
        nonZero : list of int
            indices of parameters that cannot be zero.
            they are replaced by self.tiny
        kwargs
            for internal use.
        """
        super( BaseModel, self ).__init__( )

        if copy is None :
            setatt( self, "npbase", nparams )
            setatt( self, "ndim", ndim )
            setatt( self, "priors", None )
            setatt( self, "posIndex", numpy.asarray( posIndex, dtype=int ) )
            setatt( self, "nonZero", numpy.asarray( nonZero, dtype=int ) )
            setatt( self, "deltaP", numpy.asarray( [0.00001] ) )
            setatt( self, "tiny", 1e-20 )
            parNames = ["parameter_%d"%k for k in range( self.npbase )]
            setatt( self, "parNames", parNames )
        else :
            setatt( self, "npbase", copy.npbase )
            setatt( self, "ndim", copy.ndim )
            setatt( self, "priors", None if copy.priors is None else copy.priors.copy() )
            setatt( self, "posIndex", copy.posIndex )
            setatt( self, "nonZero", copy.nonZero )
            setatt( self, "deltaP", copy.deltaP )
            setatt( self, "tiny", copy.tiny )
            setatt( self, "parNames", copy.parNames )

    def __setattr__( self, name, value ):
        """
        Set attributes.

        """
        if name == "priors" :
            setatt( self, name, value, type=Prior, isnone=True, islist=True )
            return
        if name == "parNames" :
            setatt( self, name, value, type=str, islist=True )
            return
        if name == "deltaP" :
            setatt( self, name, value, type=float, islist=True )
            return
        if name == "tiny" :
            setatt( self, name, value, type=float )
            return
        if name in ["posIndex", "nonZero"] :
            setatt( self, name, value, type=int, islist=True )
            return

        raise AttributeError(
            "Model has no attribute " + name + " of type " + str( value.__class__ ) )

    def __getattr__( self, name ) :
        """
        Return value belonging to attribute with name.

        Parameters
        ----------
        name : string
            name of the attribute
        """
        if name == "priors" :
            return None

        elif name == "parNames" :
#        if name == "parNames" :
            raise AttributeError( "Model has no defined parameter names." )

        for k,pn in enumerate( self.parNames ) :
            if name == pn :
                return self.parameters[k]

        ## Raise error for any remaining attributes, not found standardly
        raise AttributeError( "Model has no attribute %s." % name )


    #  *****RESULT**************************************************************
    def result( self, xdata, param ):
        """
        Returns the result calculated at the xdatas.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        param : array_like
            values for the parameters.

        """
        self.checkParameter( param )
        return self.baseResult( xdata, param )

    #  *****PARTIAL*************************************************************
    def partial( self, xdata, param, parlist=None ):
        """
        Returns the partial derivatives calculated at the inputs.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the result
        param : array_like
            values for the parameters.
        parlist : None or array_like
            indices of active parameters

        """
        self.checkParameter( param )
        return self.basePartial( xdata, param, parlist=parlist )

    def checkParameter( self, param ) :
        """
        Return parameters corrected for positivity and Non-zero.

        Parameters
        ----------
        param : array_like
            values for the parameters.

        """
        self.checkZeroParameter( param )
        self.checkPositive( param )
        return param

    def checkPositive( self, param ) :
        """
        Check parameters for positivity. Silently correct.

        Parameters
        ----------
        params : array_like
            values for the parameters

        """
        for k in self.posIndex :
            param[k] = abs( param[k] )

    def checkZeroParameter( self, param ):
        """
        Check parameters for Non-zero. Correct after one warning.

        Parameters
        ----------
        params : array_like
            values for the parameters

        """
        for k in self.nonZero :
            if param[k] == 0 :
                msg = ( ( self.shortName() + ": " + self.baseParameterName( k ) +
                            " ( =parameter[%d] ) equals zero."%k ) )
                warnings.warn( msg )
                param[k] = self.tiny

    def isDynamic( self ) :
        """
        Whether the model implements Dynamic
        """
        return False

    def isModifiable( self ) :
        """
        Whether the model implements Modifiable
        """
        return False

    #  *****TOSTRING***********************************************************
    def __str__( self ):
        """ Returns a string representation of the model.  """
        return self.baseName( )

    def shortName( self ):
        """
        Return a short version the string representation: upto first non-letter.

        """
        m = re.match( "^[a-zA-Z_]*", self.baseName() )
        return m.group(0)

    def derivative( self, xdata, param ) :
        """
        Returns the derivative of the model to xdata.

        It is a numeric derivative as the analytic derivative is not present
        in the model.

        Parameters
        ----------
        xdata : array_like
            values at which to calculate the derivative
        param : array_like
            values for the parameters. (default: model.parameters)

        """
        return self.baseDerivative( xdata, param )

    def setPrior( self, k, prior=None, **kwargs ) :
        """
        set the prior and/or limits for the indicated parameter.

        The prior (by default UniformPrior) is appended when k is equal to np, the
        length of the existing list of priors. It replaces the prior when k < np
        and it generates an error when k > np

        Parameters
        ---------
        k : int
            parameter number.
        prior : Prior
            prior for the parameter
        kwargs : dict
            attributes to be passed to the prior

        """
        np = Tools.length( self.priors )
        assert k <= np, "The par number %d is larger than the length of the priors %d"%(k,np)


        if prior is None :
            prior = UniformPrior( ) if np == 0 else self.basePrior( k ).copy()

        prior.setAttributes( **kwargs )

        if k == np :
            self.priors = [prior] if self.priors is None else self.priors + [prior]
            return

        self.priors[k] = prior

        return

    def hasPriors( self, isBound=True ) :
        """
        Return True when the model has priors for all its parameters.

        Parameters
        ----------
        isBound : bool
            Also check if the prior is bound.
        """
        return ( self.priors is not None and ( ( not isBound ) or
                 all( [p.isBound() for p in self.priors] ) ) )

    def getPrior( self, k ) :
        """
        Return the prior of the indicated parameter.

        Parameters
        ---------
        k : int
            parameter number.
        """
        return self.basePrior( k )

    def basePrior( self, k ) :
        """
        Return the prior of the indicated parameter.

        Parameters
        ----------
        k : int
            parameter number.
        """
        np = Tools.length( self.priors )
        if np == 0 :
            raise IndexError( "The model does not have priors." )

        if k < np:
            return self.priors[k]
        else :
            return self.priors[-1]

    def getParameterName( self, k ) :
        """
        Return the name of the indicated parameter.

        Parameters
        ---------
        k : int
            parameter number.
        """
        return self.baseParameterName( k )


    def baseParameterName( self, k ) :
        """
        Return the name of the indicated parameter.

        Parameters
        ---------
        k : int
            parameter number.
        """
        return self.parNames[k]

    def getParameterUnit( self, k ) :
        """
        Return the unit of the indicated parameter.

        Parameters
        ---------
        k : int
            parameter number.
        """
        return self.baseParameterUnit( k )

    def baseParameterUnit( self, k ) :
        """
        Return the name of the indicated parameter.

        Parameters
        ---------
        k : int
            parameter number.
        """
        return units.Unit( 1.0 )

    def hasLimits( self, fitindex=None ) :
        """
        Return True if the model has limits set.
        """
        if self.npbase == 0 :
            return True
        if self.priors is None :
            return False
        if fitindex is None :
            return all( p.hasLimits() for p in self.priors )
        else :
            haslim = True
            npr = len( self.priors )
            for k in fitindex :
                if k >= npr :
                    return haslim
                haslim = haslim and self.priors[k].hasLimits()
            return haslim

